CHARACTER_VOICE_PROFILES = {
    '': {
        'zh-CN': {
            'voiceName': 'zh-CN-YunxiNeural',
            'kwargs': {}
        },
        'en-US': {
            'voiceName': 'en-US-AndrewMultilingualNeural',
            'kwargs': {}
        }
    },
    'eilly': {
        'zh-CN': {
            'voiceName': 'zh-CN-XiaoxiaoNeural',
            'kwargs': {
                'style': 'affectionate',
            }
        },
        'en-US': {
            'voiceName': 'en-US-AriaNeural',
            'kwargs': {
                'style': 'cheerful',
            }
        }
    },
    'boy': {
        'zh-CN': {
            'voiceName': 'zh-CN-YunxiNeural',
            'kwargs': {
                'role': 'Boy',
                'style': 'Default',
                'prosody': {
                    'rate': '+4.00%'
                }
            }
        },
        'en-US': {
            'voiceName': 'zh-CN-YunxiNeural',
            'kwargs': {
                'role': 'Boy',
                'style': 'chat'
            }
        }
    },
    'girl': {
        'zh-CN': {
            'voiceName': 'zh-CN-XiaoshuangNeural',
            'kwargs': {
                'prosody': {
                    'rate': '+8.00%'
                }
            }
        },
        'en-US': {
            'voiceName': 'en-US-AnaNeural',
            'kwargs': {
                'style': 'chat'
            }
        }
    },
    'M': {
        'zh-CN': {
            'voiceName': 'zh-CN-YunxiNeural',
            'kwargs': {}
        },
        'en-US': {
            'voiceName': 'en-US-AndrewMultilingualNeural',
            'kwargs': {
                'style': 'e-learning'
            }
        }
    },
    'F': {
        'zh-CN': {
            'voiceName': 'zh-CN-XiaohanNeural',
            'kwargs': {}
        },
        'en-US': {
            'voiceName': 'en-US-JennyNeural',
            'kwargs': {
                'style': 'newscast'
            }
        }
    }
}

PRONUNCIATION_DICTIONARY = {
    "zh-CN": {
        '羞臊': '羞sào',
        '数数': '暑shù',
        '∠': '角',
        '°': '度',
        '<ul>': '',
        '</ul>': '',
        '<li>': '',
        '</li>': '。',
        '·': '-',
        '>': '大于',
        '<': '小于',
        '>=': '大于等于',
        '<=': '小于等于',
        'WHO': 'WH欧',
        'elif': 'L-If',
        'for': 'four',
    },
    "en-US": {
        '<ul>': '',
        '</ul>': '',
        '<li>': '',
        '</li>': '。',
    }
}

CHARACTER_FIGURE_ACCESSORY_KEYS = ['mask', 'cap']

STORY_SCENARIO_STYLES = {
    "shinkai_makoto": {
        "scenarios": {
            "cover": {"bgColor": "#C1D0BA"},
            "classroom": "475dc094-e147-478e-a1cf-770df5efa081",
            "notes": {
                "scene":"ce3ae229-7e6b-4e10-9d0a-e5e53d923d61",
                "htmlTemplate":"<h2 style=\'color:white\'><br/><ul><li></li></ul></h2>",
                "rect": [0.2, 0.2, 0.6, 0.8]
            },
            "exam": "a65ed1dc-9e27-4c3d-a5f7-090fc2af4917",
            "blackboard": {"bgColor": "#98A698"},
            "concentrak": {"index": "0bd6c33e-31eb-4083-8dd8-ee07837bc975", "bgColor": "#C1D0BA"},
        },
        "frame": "10px solid #843C0C",
        "positions": {
            "left": [0.2, 0.05],
            "right": [0.8, 0.05],
            "right-bottom": [0.85, -0.05],
        },
        "popup": 11,
        "transform": None
    },
    "close_up": {
        "scenarios": {
            "cover": "5cdab8ba-c028-45ea-87cc-b0d75dd81e10",
            "classroom": "b99b0dc2-0387-4a4f-9c61-308b362cb8f6",
            "notes": {
                "scene": "6c5c45c3-070f-439a-a6f7-a366a04c357c",
                "htmlTemplate":"<p><ul style=\'color: white\' size='5'><li></li></ul></p>",
                "rect": [0.1, 0.28, 0.7, 0.5]
            },
            "exam": "ea874992-c7a9-44c4-8bef-13cad8cee917",
            "blackboard": "e183c517-cbb7-4831-bcf6-efd310ee8790",
            "concentrak": {"index": "0bd6c33e-31eb-4083-8dd8-ee07837bc975", "bgColor": "#9BB9BF"},
        },
        "frame": "10px solid #843C0C",
        "positions": {
            "left": [0.2, -0.25],
            "right": [0.8, -0.25],
            "right-bottom": [0.8, 0.05],
        },
        "popup": 10,
        "transform": "scale(2)"
    }
}


import re
from bs4 import BeautifulSoup
import xml.etree.ElementTree as ET
from PIL import Image
import os

VISIBLE_ACTORS=("boy", "girl", "eily", "eilly")
INVISIBLE_ACTORS=("", "M", "F")
SCENARIO_ACTORS=("ending", "exam", "concentrak", "notes")

def remove_emojis(text):
    emojiPattern = re.compile(pattern = "["
        u"\U0001F600-\U0001F64F"  # emoticons
        u"\U0001F300-\U0001F5FF"  # symbols & pictographs
        u"\U0001F680-\U0001F6FF"  # transport & map symbols
        u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
        u"\U0001F926-\U0001F991"
                    "]+", flags = re.UNICODE)
    return re.sub(emojiPattern, '', text)

def has_chinese_char(text):
  """Checks if a string contains at least one Chinese character.

  Args:
      text: The string to be checked.

  Returns:
      True if the string contains at least one Chinese character, False otherwise.
  """
  # Check if any character in the string falls within the Unicode range of Chinese characters
  return any(u'\u4e00' <= char <= u'\u9fff' for char in text)

def update_object(original, update, default_locale=None):
    """Updates the original object from the update object.

    Args:
        original: The object to be updated.
        update: The object containing updates.

    Returns:
        A new object with the updates applied. Ignore if update is None.
    """
    if update == None:
        return original
    
    if isinstance(update, dict):
        return {
            **(original \
                if isinstance(original, dict) \
                else ({} if original == None else {default_locale: original})), \
            **{key: update[key] for key in update}
        }
    elif isinstance(original, dict) and default_locale != None:
        original[default_locale] = update
    else:
        original = update
    return original

def get_actors(objects):
    assert isinstance(objects, list)
    actor = None
    narrator = None
    defaultObject = None
    actorId = -1
    narratorId = -1
    defaultObjectId = -1
    for i, object in enumerate(objects):
        if object.get("name", None) in VISIBLE_ACTORS:
            actor = object["name"]
            actorId = i
        elif object.get("name", None) in INVISIBLE_ACTORS:
            narrator = object["name"]
            narratorId = i
        else:
            defaultObject = object.get("name", None)
            defaultObjectId = i
    return actor, actorId, narrator, narratorId, defaultObject, defaultObjectId

def update_visible_actor(objects, actor):
    assert isinstance(objects, list) and actor in VISIBLE_ACTORS
    if len(objects) == 0:
        objects.append({"name": actor})
    else:
        for i, object in enumerate(objects):
            if object["name"] in VISIBLE_ACTORS:
                objects[i]["name"] = actor

def update_invisible_actor(objects, actor):
    assert isinstance(objects, list) and actor in INVISIBLE_ACTORS
    if len(objects) == 0:
        objects.append({"name": actor})
    else:
        for i, object in enumerate(objects):
            if object["name"] in INVISIBLE_ACTORS:
                objects[i]["name"] = actor

def switch_to_test_path(path):
    if path.startswith("/story/"):
        return "/test/" + path[len("/story/"):]
    else:
        return path

def switch_to_basename(path):
    return os.path.basename(path)

def reset_voices_to_basename(scriptList, oldNarrator, newNarrator):
    assert isinstance(scriptList, list) \
        and (oldNarrator in VISIBLE_ACTORS or oldNarrator in INVISIBLE_ACTORS) \
        and (newNarrator in VISIBLE_ACTORS or newNarrator in INVISIBLE_ACTORS)
    for i, script in enumerate(scriptList):
        if "narrator" in script and script["narrator"] == oldNarrator and "sound" in script \
            and isinstance(script["sound"], str) and len(script["sound"]) > 0:
            scriptList[i]["narrator"] = newNarrator
            scriptList[i]["sound"] = switch_to_basename(script["sound"])
            scriptList[i].pop("languages", None)

def get_image_from_board(boardObject):
    image = None
    rect = None
    caption = None
    if "content" in boardObject:
        rect = boardObject["rect"]
        image = boardObject["content"].get("image", None)
        video = boardObject["content"].get("src", None)
        videoType = boardObject["content"].get("videoType", None)
        caption = boardObject["content"].get("caption", None)
    return rect, image, video, videoType, caption

def get_html_from_board(boardObject):
    html = None
    rect = boardObject.get("rect", None)
    if "content" in boardObject:
        html = boardObject["content"].get("html", None)
    return rect, html

def get_question_from_board(boardObject):
    question = None
    options = None
    answer = None
    rect = None
    colsPerRow = 1
    fontSize = 20
    fontColor = "white"
    rect = boardObject.get("rect", rect)
    if "content" in boardObject:
        question = boardObject["content"].get("question", question)
        options = boardObject["content"].get("options", options)
        answer = boardObject["content"].get("answer", answer)
        colsPerRow = boardObject["content"].get("colsPerRow", colsPerRow)
        fontSize = boardObject["content"].get("fontSize", fontSize)
        fontColor = boardObject["content"].get("fontColor", fontColor)
    return rect, question, options, answer, colsPerRow, fontSize, fontColor

def get_subscript_from_interaction(interactionObject):
    actor = -1
    voice = -1
    figure = -1
    text = None
    duration = ""
    if "content" in interactionObject:
        text = interactionObject["content"].get("text", text)
        voice = interactionObject["content"].get("voice", voice)
        actor = interactionObject.get("actor", actor)
        figure = interactionObject.get("figure", figure)
        duration = interactionObject.get("duration", duration)
    return actor, figure, text, voice, duration

def get_bullets_from_html(html:str):
    # Define a pattern to match the content within list items (ul or li tags)
    pattern = r"<(ul|li)>(.*?)</(ul|li)>"

    # Extract content from html using findall
    matches = re.findall(pattern, html, flags=re.DOTALL)

    extracted = []
    if matches:
        # Remove tags using re.sub
        extracted = [re.sub(r"<[^>]+>", "", match[1]) for match in matches]
    return extracted

# Extract all text recursively, excluding script and style tags
def retrieve_element_text(element):
    text_object_list = []
    if element.name:
        for child in element.children:
            text_object_list.extend(retrieve_element_text(child))
    elif element.string:
        text_object_list.append({element.parent.name: element.string.strip()})  # Add current element's text
    return text_object_list

def extract_html_elements(html_text:str):
    soup = BeautifulSoup(html_text, "lxml")
    
    # Extract all text into a list
    extracted_text_object_list = retrieve_element_text(soup)
    
    # Optional: Flatten the list of lists (if nested due to structure)
    if any(isinstance(item, list) for item in extracted_text_object_list):
        extracted_text_object_list = [item for sublist in extracted_text_object_list for item in sublist]

    html_template = html_text
    for string in extracted_text_object_list:
        key = next(iter(string), None)
        if key != None:
            html_template = html_template.replace(">"+string[key]+"<", ">{"+key+"}<")
    
    # Regular expression pattern to match any number of continuous <li> elements
    # Explanation of the pattern:
    # - "<li>(.*?)</li>": Matches the first <li> tag and its content (non-greedy)
    # - "(?:<li>(.*?)</li>)*": Matches zero or more repetitions of subsequent <li> tags and their content (non-greedy)
    pattern = r"<li>(.*?)</li>(?:<li>(.*?)</li>)*"
        
    # Replace each group with a single '{}'
    html_template = re.sub(pattern, "<li>{li}</li>", html_template)

    return extracted_text_object_list, html_template

def retrieve_svg_size(image_path):
    # Load the SVG file
    tree = ET.parse(image_path)
    root = tree.getroot()

    # Extract attributes from the <svg> tag
    width = root.get("width", 0)  # Get the width attribute
    height = root.get("height", 0)  # Get the height attribute
    viewBox = root.get("viewBox", "0, 0, 0, 0")  # Get the viewBox attribute

    split_pattern = r"[ ,]+"

    return [int(width), int(height)], [
        int(float(num)) for num in re.split(split_pattern, viewBox)
    ]

def retrieve_pixel_size(image_path):
    width = height = 0
    try:
        # Open the image using the Python Imaging Library (PIL)
        image = Image.open(image_path)

        # Get the width and height of the image in pixels
        width, height = image.size

        image.close()
    except OSError as e:
        print(f"Error opening image: {e}")

    # Return the width and height as a tuple
    return width, height

def get_image_size(file_path):
    width = height = 0
    try:
        if ".svg" in file_path[-4:]:
            dim2, dim4 = retrieve_svg_size(file_path)
            if dim2 == [0, 0]:
                width = dim4[2]
                height = dim4[3]
            else:
                width = dim2[0]
                height = dim2[1]
        elif (
            ".jpg" in file_path[-4:]
            or ".jpeg" in file_path[-5:]
            or ".png" in file_path[-4:]
            or ".gif" in file_path[-4:]
        ):
            width, height = retrieve_pixel_size(file_path)
    except:
        print("Retrieve image size error for", file_path)
    return width, height

def is_valid_http_url(url_str):
    return url_str.startswith("https://") or url_str.startswith("http://")

def merge_dicts(dict1, dict2):
  """Merges two dictionaries recursively, handling different structures.

  Args:
      dict1: The first dictionary.
      dict2: The second dictionary.

  Returns:
      A new dictionary with merged key-value pairs.
  """

  merged_dict = dict1.copy()  # Start with a copy of the first dictionary
  for key, value in dict2.items():
    if key in merged_dict:
      if isinstance(merged_dict[key], dict) and isinstance(value, dict):
        # Recursively merge nested dictionaries
        merged_dict[key] = merge_dicts(merged_dict[key], value)
      else:
        # Overwrite existing key-value pair (non-nested case)
        merged_dict[key] = value
    else:
      # Add new key-value pair from the second dictionary
      merged_dict[key] = value
  return merged_dict


CHARACTER_FIGURES = {
    "eily": [
        "robot-girl-3123483124.svg",
    ],
    "boy": [
        "boy-stand-arms-fistup-leg-step-blinkgrin-anime",
        "boy-stand-arms-open-leg-open-question-anime", 
        "boy-stand-arms-straight-leg-straddle-calm-anime", 
        "boy-stand-arms-straight-leg-straight-blinkgrin-anime", 
        "boy-stand-arms-straight-leg-straight-blinksay-anime", 
        "boy-stand-arms-straight-leg-straight-complain-anime", 
        "boy-stand-arms-straight-leg-straight-groan-anime", 
        "boy-stand-arms-straight-leg-straight-say-anime", 
        "boy-stand-arms-straight-leg-straight-smilesay-anime", 
        "boy-stand-left-finger1up-right-straight-leg-crossing-blinkgrin-anime", 
        "boy-stand-left-finger1up-right-straight-leg-step-blinkgrin-anime", 
        "boy-stand-left-finger1up-right-straight-leg-step-groan-anime", 
        "boy-stand-left-finger1up-right-straight-leg-straddle-calm-anime", 
        "boy-stand-left-fistup-right-straight-leg-straddle-blinkgrin-anime", 
        "boy-stand-left-fistup-right-straight-leg-straddle-complain-anime", 
        "boy-stand-left-fistup-right-straight-leg-straddle-groan-anime", 
        "boy-stand-left-fistup-right-straight-leg-straddle-smilesay-anime", 
        "boy-stand-left-straight-right-handonhip-leg-straddle-calm-anime", 
        "boy-stand-left-thumbup-right-handonhip-leg-straddle-smilesay-anime", 
        "boy-stand-left-thumbup-right-handonhip-leg-straddle-winking-anime", 
        "boy-stand-left-thumbup-right-handonhip-leg-straight-calm-anime", 
        "boy-standright-left-handonhip-right-thumbup-leg-straddle-winking-anime", 
        "boy-standright-left-straight-right-finger1up-leg-straddle-winking-anime", 
        "boy-standright-left-straight-right-fistup-leg-straddle-smilesay-anime", 
        "boy-standright-left-straight-right-fistup-leg-straddle-winking-anime", 
        "boy-half-left-handonhip-right-finger1up-calm-anime", 
        "boy-half-left-handonhip-right-open-smilesay-anime", 
        "boy-half-left-handonhip-right-thumbup-winking-anime", 
        "boy-head-smilesay-anime", 
        "boy-standright-left-straight-right-finger1up-leg-crossing-blinkgrin-anime", 
        "boy-stand-arms-straight-leg-straddle-smilesay-anime", 
        "boy-stand-arms-straight-leg-straddle-groan-anime", 
        "boy-stand-arms-straight-leg-straddle-blinkgrin-anime", 
        "boy-stand-left-fistup-right-straight-leg-step-groan-anime", 
        "boy-stand-left-fistup-right-straight-leg-step-blinkgrin-anime", 
        "boy-stand-left-fistup-right-straight-leg-crossing-blinkgrin-anime", 
        "boy-half-left-handonhip-right-thumbup-smilesay-anime", 
        "boy-half-arms-straight-winking-anime", 
        "boy-stand-left-straight-right-spear-leg-straddle-khoikhoimask-anime", 
        "boy-stand-left-fistup-right-spear-leg-straddle-khoikhoimask-anime", 
        "boy-stand-left-finger1up-right-straight-leg-crossing-calm-anime"
    ], 
    "girl": [
        "girl-stand-arms-fistup-leg-open-blinkgrin-anime", 
        "girl-stand-arms-handonhip-leg-straight-calm-anime", 
        "girl-stand-arms-handonhip-leg-straight-complain-anime", 
        "girl-stand-arms-handonhip-leg-straight-smilesay-anime", 
        "girl-stand-arms-handonhip-leg-straight-winking-anime", 
        "girl-stand-arms-open-leg-straight-bemused-anime", 
        "girl-stand-arms-open-leg-straight-blinkgrin-anime", 
        "girl-stand-arms-open-leg-straight-question-anime", 
        "girl-stand-arms-open-leg-straight-smilesay-anime", 
        "girl-stand-arms-straight-leg-straight-calm-anime", 
        "girl-stand-left-handonhip-right-thumbup-leg-open-winking-anime", 
        "girl-stand-left-straight-right-finger1up-leg-crossing-question-anime", 
        "girl-stand-left-straight-right-finger1up-leg-straight-groan-anime", 
        "girl-stand-left-straight-right-finger1up-leg-straight-smilesay-anime", 
        "girl-stand-left-straight-right-finger1up-leg-straight-winking-anime", 
        "girl-stand-left-handonhip-right-open-leg-ontiptoe-groan-anime", 
        "girl-half-left-handonhip-right-finger1up-smilesay-anime", 
        "girl-half-left-handonhip-right-open-smilesay-anime", 
        "girl-half-left-handonhip-right-thumbup-winking-anime", 
        "girl-head-smilesay-anime", 
        "girl-stand-left-shield-right-open-leg-straight-khoikhoimask-anime", 
        "girl-stand-left-shield-right-finger1up-leg-straight-khoikhoimask-anime", 
        "girl-stand-arms-handonhip-leg-straight-groan-captaincap-anime", 
        "girl-stand-left-open-right-finger1up-leg-straight-smilesay-captaincap-anime", 
        "girl-stand-left-open-right-finger1up-leg-straight-winking-captaincap-anime", 
        "girl-stand-arms-open-leg-straight-captaincap-anime"
    ]
}


from typing import List, Dict, Union
import copy

from .utility import *

RED    = "\033[91m"
GREEN  = "\033[92m"
YELLOW = "\033[93m"
BLUE   = "\033[94m"
RESET  = "\033[0m"

DEFAULT_LANGUAGE="zh-CN"
LANGUAGE_ENG="en-US"
DEFAULT_NARRATOR="M"
VISIBLE_ACTORS=("boy", "girl", "eily", "eilly")
VISIBLE=0
INVISIBLE_ACTORS=("", "M", "F")
INVISIBLE=1
SCENARIO_ACTORS=("ending", "exam", "concentrak", "notes")
SCENARIO=2
LOCAL_DEFAULT_ROOT="./test"

BREAK_TIME="<break time=\"1500ms\"/>"
BULLET_KEY="li"


ENDING_SOUND="/story/audios/OurMusicBox - 24 Hour Coverage - intro.mp3"

class MText:
    def __init__(self, value:Union[str, dict[str, str]], language:str=None):
        if isinstance(value, MText):
            self.data = copy.deepcopy(value.data)
        elif isinstance(value, str):
            self.data = value if language == None else {language: value}
        elif isinstance(value, dict) and any(isinstance(v, str) for v in value.values()):
            self.data = {}
            for key in value:
                if isinstance(value[key], str):
                    self.data[key] = value[key]
        else:
            self.data = ""
    
    def len(self):
        return len(self.data)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.data == other
        elif isinstance(other, MText):
            return self.data == other.data
        return False        
      
    def copy(self):
        return copy.deepcopy(self)
    
    def firstValidPair(self):
        if isinstance(self.data, dict):
            key = next(iter(self.data), None)
            return key, self.data[key] if key != None else None
        else:
            return None, self.data
    
    def merge(self, newData):
        if isinstance(self.data, str):
            if isinstance(newData, dict):
                result = MText({**{DEFAULT_LANGUAGE:self.data}, **newData})
            elif isinstance(newData, MText):
                result = MText(self.data, DEFAULT_LANGUAGE).merge(newData.data)
            elif newData != None and len(newData) > 0:
                result = MText(newData)
            else:
                result = self.copy()
        else:
            if isinstance(newData, dict):
                result = MText({**self.data, **newData})
            elif isinstance(newData, MText):
                result = self.merge(newData.data)
            elif newData != None and len(newData) > 0:
                result = MText({**self.data, **{DEFAULT_LANGUAGE:newData}})
            else:
                result = self.copy()
        return result

    def export(self):
        if isinstance(self.data, str):
            return self.data
        elif isinstance(self.data, dict):
            data = {}
            for key in self.data.keys():
                if isinstance(self.data[key], str) and len(self.data[key])>0:
                    data[key] = self.data[key]
            if len(data) > 0:
                return data
            return None

class MList:
    def __init__(self, value:Union[str, list, dict[str, list]], language:str=None):
        if isinstance(value, str):
            self.data = [value] if language == None else {language: [value]}
        elif isinstance(value, list):
            self.data = value if language == None else {language: value}
        elif isinstance(value, dict) and any(isinstance(v, list) for v in value.values()):
            self.data = {}
            for key in value:
                if isinstance(value[key], list):
                    self.data[key] = value[key]
        elif isinstance(value, MList):
            self.data = copy.deepcopy(value.data)
        else:
            self.data = []

    def __eq__(self, other):
        if isinstance(other, list):
            return self.data == other
        elif isinstance(other, MList):
            return self.data == other.data
        return False

    def copy(self):
        return copy.deepcopy(self)
    
    def getMTextByPos(self, pos):
        if pos < len(self.firstValidPair()[1]) and pos >= 0:
            if isinstance(self.data, dict):
                result = {}
                for key in self.data:
                    result[key] = self.data[key][pos]
                return MText(result)
            return MText(self.data[pos])
        
    def getListByLanguage(self, language):
        if isinstance(self.data, dict) and language in self.data.keys():
            return copy.deepcopy(self.data[language])
        elif isinstance(self.data, list) and language == DEFAULT_LANGUAGE:
            return copy.deepcopy(self.data)
        return []

    def firstValidPair(self):
        if isinstance(self.data, dict):
            key = next(iter(self.data), None)
            return key, self.data[key] if key != None else None
        else:
            return None, self.data

    def export(self):
        if isinstance(self.data, list):
            return self.data.copy()
        elif isinstance(self.data, dict):
            data = {}
            for key in self.data.keys():
                if isinstance(self.data[key], list):
                    data[key] = self.data[key]
            if len(data) > 0:
                return data
            return None
        
class HTML:
    def __init__(self, template:str=None, textList:List[Dict[str, str]]=None):
        self.template = template
        self.bullets = []
        self.elements = []
        if isinstance(textList, list):
            for textObject in textList:
                if isinstance(textObject, dict):
                    key = next(iter(textObject), None)
                    if key != None and key == BULLET_KEY: #bullet tag
                        self.bullets.append(textObject[key])
                    elif key != None:
                        self.elements.append(textObject)

    @staticmethod
    def loadFromText(htmlText:str):
        textList, template = extract_html_elements(htmlText)
        return HTML(template=template, textList=textList)
    
    def copy(self):
        return copy.deepcopy(self)

    def setTemplate(self, template):
        self.template = template

    def addElement(self, tag, text):
        self.elements.append({tag: text})

    def updateElement(self, pos, tag, text):
        if pos < len(self.elements) and pos >= 0:
            self.elements[pos] = {tag: text}

    def removeElement(self, pos):
        if pos < len(self.elements) and pos >= 0:
            self.elements.pop(pos)

    def addBullet(self, text):
        self.bullets.append(text)

    def updateBullet(self, pos, text):
        if pos < len(self.bullets) and pos >= 0:
            self.bullets[pos] = text

    def removeBullet(self, pos):
        if pos < len(self.bullets) and pos >= 0:
            self.bullets.pop(pos)

    def exportBullets(self):
        return self.bullets.copy()

    def exportElements(self):
        return self.elements.copy()

    def export(self):
        result = self.template
        if isinstance(result, str):
            for entry in self.elements:
                key = next(iter(entry), None)
                if key != None:
                    result = result.replace("{"+key+"}", entry[key], 1)
            result = result.replace("{"+BULLET_KEY+"}", ("</"+BULLET_KEY+"><"+BULLET_KEY+">").join(self.bullets))
        return result
    
    def exportScripts(self):
        if isinstance(self.template, str):
            resultList = []
            outScripts = copy.deepcopy(self.elements)
            pattern = r"\{([a-zA-Z0-9]+)\}"  # Matches { followed by alphanumeric characters
            # Find all occurrences and extract the placeholder names
            placeholders = [match.group(1) for match in re.finditer(pattern, self.template)]
            for placeholder in placeholders:
                if placeholder == BULLET_KEY:
                    resultList.append(BREAK_TIME.join(self.bullets))
                else:
                    for j, pair in enumerate(outScripts):
                        if next(iter(pair), None) == placeholder:
                            resultList.append(next(iter(pair.values()), None))
                            outScripts.pop(j)
                            break

            return BREAK_TIME.join(resultList)

        return None
        
class MHTML:
    def __init__(self, template:str, data):
        self.template = template
        self.data = data
    
    @staticmethod
    def loadFromText(htmlText:MText):
        if isinstance(htmlText, MText):
            htmlObject = htmlText.export()
        elif isinstance(htmlText, dict) or isinstance(htmlText, str):
            htmlObject = htmlText
        else:
            return MHTML(None, None)
        template = None
        if isinstance(htmlObject, str):
            htmlData = HTML.loadFromText(htmlObject)
            return MHTML(htmlData.template, htmlData)
        else:
            htmlDict = {}
            for key in htmlObject:
                htmlDict[key] = HTML.loadFromText(htmlObject[key])
            if DEFAULT_LANGUAGE not in htmlObject:
                print(f"{YELLOW}WARNING{RESET}: No language version for {DEFAULT_LANGUAGE} detected, switch to first available language")
                template = htmlDict[next(iter(htmlDict), None)].template
            else:
                template = htmlDict[DEFAULT_LANGUAGE].template

            return MHTML(template, htmlDict)
    
    def copy(self):
        return copy.deepcopy(self)
    
    def setTemplate(self, template:str):
        if isinstance(self.data, HTML):
            self.data.setTemplate(template)
        elif isinstance(self.data, dict):
            data = {}
            for key in self.data:
                data[key] = self.data[key].setTemplate(template)

    def addElement(self, tag, text, **kwargs):
        if tag == BULLET_KEY:
            self.addBullet(text)
        else:
            if isinstance(text, str):
                if isinstance(self.data, HTML):
                    self.data.addElement(tag, text)
                elif isinstance(self.data, dict):
                    if kwargs.get("targetLanguage", None) != None:
                        self.data[kwargs["targetLanguage"]].addElement(tag, text)
                    else:
                        for key in self.data:
                            self.data[key].addElement(tag, text)
            elif isinstance(text, dict):
                if isinstance(self.data, HTML):
                    self.data = {DEFAULT_LANGUAGE: self.data}
                if isinstance(self.data, HTML) or isinstance(self.data, dict):
                    for key in text:
                        if key not in self.data:
                            self.data[key] = self.data[DEFAULT_LANGUAGE]
                        for key in self.data:
                            self.addElement(tag, text[key] if key in text else text[next(iter(text), None)], targetLanguage=key)
            elif isinstance(text, MText):
                self.addElement(tag, text.data)

    def updateElement(self, pos, tag, text, **kwargs):
        if pos < len(self.data.bullets \
                     if isinstance(self.data, HTML) \
                     else (self.data[DEFAULT_LANGUAGE].bullets \
                            if isinstance(self.data, dict) \
                            else [])) and pos >= 0:
            self.data.bullets[pos] = text

            if isinstance(text, str):
                if isinstance(self.data, HTML):
                    self.data.updateElement(pos, tag, text)
                elif isinstance(self.data, dict):
                    if kwargs.get("targetLanguage", None) != None:
                        self.data[kwargs["targetLanguage"]].updateElement(pos, tag, text)
                    else:
                        for key in self.data:
                            self.data[key].updateElement(pos, tag, text)
            elif isinstance(text, dict):
                if isinstance(self.data, HTML):
                    self.data = {DEFAULT_LANGUAGE: self.data}
                if isinstance(self.data, HTML) or isinstance(self.data, dict):
                    for key in text:
                        if key not in self.data:
                            self.data[key] = self.data[DEFAULT_LANGUAGE]
                        for key in self.data:
                            self.updateElement(pos, tag, text[key] if key in text else text[next(iter(text), None)], targetLanguage=key)
            elif isinstance(text, MText):
                self.updateElement(pos, tag, text.data)
        
            return True
        
        return False

    def removeElement(self, pos):
        if pos < len(self.data.bullets \
                     if isinstance(self.data, HTML) \
                     else (self.data[DEFAULT_LANGUAGE].bullets \
                            if isinstance(self.data, dict) \
                            else [])) and pos >= 0:
            if isinstance(self.data, HTML):
                self.data.removeElement(pos)
            elif isinstance(self.data, dict):
                for key in self.data:
                    self.data[key].removeElement(pos)
            
            return True
        
        return False

    def addBullet(self, text, **kwargs):
        if isinstance(text, str):
            if isinstance(self.data, HTML):
                self.data.addBullet(text)
            elif isinstance(self.data, dict):
                if kwargs.get("targetLanguage", None) != None:
                    self.data[kwargs["targetLanguage"]].addBullet(text)
                else:
                    for key in self.data:
                        self.data[key].addBullet(text)
        elif isinstance(text, dict):
            if isinstance(self.data, HTML):
                self.data = {DEFAULT_LANGUAGE: self.data}
            if isinstance(self.data, HTML) or isinstance(self.data, dict):
                for key in text:
                    if key not in self.data:
                        self.data[key] = self.data[DEFAULT_LANGUAGE]
                    for key in self.data:
                        self.addBullet(text[key] if key in text else text[next(iter(text), None)], targetLanguage=key)
        elif isinstance(text, MText):
            self.addBullet(text.data)

    def updateBullet(self, pos, text, **kwargs):
        if pos < len(self.data.bullets \
                     if isinstance(self.data, HTML) \
                     else (self.data[DEFAULT_LANGUAGE].bullets \
                            if isinstance(self.data, dict) \
                            else [])) and pos >= 0:
            self.data.bullets[pos] = text

            if isinstance(text, str):
                if isinstance(self.data, HTML):
                    self.data.updateBullet(pos, text)
                elif isinstance(self.data, dict):
                    if kwargs.get("targetLanguage", None) != None:
                        self.data[kwargs["targetLanguage"]].updateBullet(pos, text)
                    else:
                        for key in self.data:
                            self.data[key].updateBullet(pos, text)
            elif isinstance(text, dict):
                if isinstance(self.data, HTML):
                    self.data = {DEFAULT_LANGUAGE: self.data}
                if isinstance(self.data, HTML) or isinstance(self.data, dict):
                    for key in text:
                        if key not in self.data:
                            self.data[key] = self.data[DEFAULT_LANGUAGE]
                        for key in self.data:
                            self.updateBullet(pos, text[key] if key in text else text[next(iter(text), None)], targetLanguage=key)
            elif isinstance(text, MText):
                self.updateBullet(pos, text.data)
        
            return True
        
        return False

    def removeBullet(self, pos):
        if pos < len(self.data.bullets \
                     if isinstance(self.data, HTML) \
                     else (self.data[DEFAULT_LANGUAGE].bullets \
                            if isinstance(self.data, dict) \
                            else [])) and pos >= 0:
            if isinstance(self.data, HTML):
                self.data.removeBullet(pos)
            elif isinstance(self.data, dict):
                for key in self.data:
                    self.data[key].removeBullet(pos)
            
            return True
        
        return False

    def exportBullets(self):
        if isinstance(self.data, HTML):
            return self.data.exportBullets()
        elif isinstance(self.data, dict):
            data = {}
            for key in self.data:
                data[key] = self.data[key].exportBullets()
            return data
        return None

    def exportElements(self):
        if isinstance(self.data, HTML):
            return self.data.exportElements()
        elif isinstance(self.data, dict):
            data = {}
            for key in self.data:
                data[key] = self.data[key].exportElements()
            return data
        return None

    def export(self):
        if isinstance(self.data, HTML):
            return self.data.export()
        elif isinstance(self.data, dict):
            data = {}
            for key in self.data:
                data[key] = self.data[key].export()
            return data
        return None
    
    def exportScripts(self):
        if isinstance(self.data, HTML):
            return self.data.exportScripts()
        elif isinstance(self.data, dict):
            data = {}
            for key in self.data:
                data[key] = self.data[key].exportScripts()
            return data
        return None

class Content:
    def __init__(self, popup:int=None, voice:int=None, text:MText=None, textAlign:str=None, image:MText=None, src:MText=None, videoType:str=None, \
                 magnify:bool=None, caption:MText=None, fontSize:Union[str, int]=None, fontColor:str=None, colsPerRow:int=None, border:str=None, \
                 rect:list=None, question:MText=None, options:MList=None, answer:Union[str, MList]=None, html:MHTML=None, type:str=None):
        self.popup = popup
        self.voice = voice
        self.text = MText(text) if text !=None else None
        self.textAlign = textAlign
        self.image = MText(image) if image !=None else None
        self.src = MText(src) if src!=None else None
        self.videoType = videoType
        self.magnify = magnify
        self.caption = MText(caption) if caption!=None else None
        self.fontSize = fontSize if (isinstance(fontSize, str) and fontSize.endswith("px")) else \
            str(fontSize)+"px" if isinstance(fontSize, int) else None
        self.fontColor = fontColor
        self.colsPerRow = colsPerRow
        self.border = border
        self.rect = rect[:4] if (isinstance(rect, list) and len(rect)>=4) else None
        self.question = MText(question) if question!=None else None
        self.options = MList(options) if options!=None else None
        self.answer = MList(answer) if answer!=None else None
        self.html = html
        self.type = type

    @staticmethod
    def load(object):
        if isinstance(object, dict):
            return Content(
                popup = object.get("popup", None),
                voice = object.get("voice", None),
                text = MText(object["text"]) if object.get("text", None) != None else None,
                textAlign = object.get("textAlign", None),
                image = MText(object["image"]) if object.get("image", None) != None else None,
                src = MText(object["src"]) if object.get("src", None) != None else None,
                videoType = object.get("videoType", None),
                magnify = object.get("magnify", None),
                caption = MText(object["caption"]) if object.get("caption", None) != None else None,
                fontSize = object.get("fontSize", None),
                fontColor = object.get("fontColor", None),
                colsPerRow = object.get("colsPerRow", None),
                border = object.get("border", None),
                rect = object.get("rect", None),
                question = MText(object["question"]) if object.get("question", None) != None else None,
                options = MList(object["options"]) if object.get("options", None) != None else None,
                answer = MList(object["answer"]) if object.get("answer", None) != None else None,
                html = MHTML.loadFromText(MText(object["html"])) if object.get("html", None) != None else None,
                type = object.get("type", None)
            )
        return None

    def copy(self):
        return copy.deepcopy(self)

    def exportScripts(self):
        data = {"transcripts":[]}
        if self.voice!=None:
            data["voice"] = self.voice
        if self.text!=None and self.text.export()!=None:
            data["transcripts"].append(self.text.export())
        if self.caption!=None and self.caption.export()!=None:
            data["transcripts"].append(self.caption.export())
        if self.question!=None and self.question.export()!=None:
            data["transcripts"].append(self.question.export())
        if self.options!=None and self.options.export()!=None:
            for i in range(len(self.options.firstValidPair()[1])):
                data["transcripts"].append(self.options.getMTextByPos(i))
        if self.answer!=None and self.answer.export()!=None:
            for i in range(len(self.answer.firstValidPair()[1])):
                data["transcripts"].append(self.answer.getMTextByPos(i))
        if self.html!=None and self.html.export()!=None:
            data["transcripts"].append(self.html.export())

        if len(data["transcripts"]) > 0:
            return data
        
        return None

    def export(self):
        data = {}
        if self.popup!=None:
            data["popup"] = self.popup
        if self.voice!=None:
            data["voice"] = self.voice
        if self.text!=None and self.text.export()!=None:
            data["text"] = self.text.export()
        if self.textAlign!=None:
            data["textAlign"] = self.textAlign
        if self.image!=None and self.image.export()!=None:
            data["image"] = self.image.export()
        if self.src!=None and self.src.export()!=None:
            data["src"] = self.src.export()
        if self.videoType!=None:
            data["videoType"] = self.videoType
        if self.magnify==True:
            data["magnify"] = self.magnify
        if self.caption!=None and self.caption.export()!=None:
            data["caption"] = self.caption.export()
        if self.fontSize!=None:
            data["fontSize"] = self.fontSize
        if self.fontColor!=None:
            data["fontColor"] = self.fontColor
        if self.colsPerRow!=None:
            data["colsPerRow"] = self.colsPerRow
        if self.border!=None:
            data["border"] = self.border
        if self.rect!=None:
            data["rect"] = self.rect
        if self.question!=None and self.question.export()!=None:
            data["question"] = self.question.export()
        if self.options!=None and self.options.export()!=None:
            data["options"] = self.options.export()
        if self.answer!=None and self.answer.export()!=None:
            data["answer"] = self.answer.export()
        if self.html!=None and self.html.export()!=None:
            data["html"] = self.html.export()
        if self.type!=None:
            data["type"] = self.type

        if len(data) > 0:
            return data
        
        return None

class Board:
    def __init__(self, content:Content=None, type:str=None, rect:list=None, contentList:list=None):
        self.content=content if isinstance(content, Content) else Content()
        self.type=type
        self.rect=rect[:4] if (isinstance(rect, list) and len(rect)>=4) else None
        self.contentList = copy.deepcopy(contentList) if isinstance(contentList, list) else []

    @staticmethod
    def load(object):
        if isinstance(object, dict):
            contentList = []
            if isinstance(object.get("contentList", None), list):
                for content in object["contentList"]:
                    contentList.append(Content.load(content))
            return Board(
                content = Content.load(object.get("content", None)),
                type = object["type"] \
                    if object.get("type", None)!=None \
                    else object.get("content", {}).get("type", None),
                rect = object["rect"] \
                    if object.get("rect", None)!=None \
                    else object.get("content", {}).get("rect", None),
                contentList = contentList if len(contentList)>0 else None
            )
        return Board()
        
    def copy(self):
        return copy.deepcopy(self)

    def export(self):
        data = {}

        if isinstance(self.content, Content) and self.content.export()!=None:
            data["content"] = self.content.export()
        if self.type!=None:
            data["type"] = self.type
        if self.rect!=None:
            data["rect"] = self.rect
        if self.contentList!=None and len(self.contentList) > 0:
            data["contentList"] = []
            for i, content in enumerate(self.contentList):
                if isinstance(content, Content) and content.export() != None:
                    data["contentList"].append(content.export())
                else:
                    print(f"Invalid Content instance {content} at index {i}")
        return data
    
class Scene:
    def __init__(self, value:str=None, index:str=None, bgColor:str=None):
        self.value = value
        self.index = index
        self.bgColor = bgColor

    @staticmethod
    def load(object):
        if isinstance(object, str):
            return Scene(value = object)
        elif isinstance(object, dict):
            return Scene(
                index = object.get("index", None),
                bgColor = object.get("bgColor", None)
                )
        else:
            return Scene()

    def copy(self):
        return copy.deepcopy(self)
    
    def export(self):
        result = ""
        if isinstance(self.value, str):
            result = self.value
        else:
            data = {}
            if isinstance(self.index, str) and len(self.index) > 0:
                data["index"] = self.index
            if isinstance(self.bgColor, str) and len(self.bgColor) > 0:
                data["bgColor"] = self.bgColor
            if len(data) > 0:
                result = data
    
        return result

class Script:
    def __init__(self, transcript:MText=None, sound:str=None, narrator:str=None, alternative:MText=None, languages:list=None, soundReady:bool=False):
        self.transcript = MText(transcript) if transcript!=None else None
        self.sound = sound
        self.narrator = narrator
        self.alternative = MText(alternative) if alternative!=None else None
        self.soundReady = bool(soundReady)
        self._languages = []
        if isinstance(languages, list) or isinstance(languages, tuple):
            for language in languages:
                if isinstance(language, str) and len(language) > 0:
                    self._languages.append(language)
        elif isinstance(languages, str) and len(languages) > 0:
            self._languages = [languages]
        
    def copy(self):
        return copy.deepcopy(self)

    def reset2basename(self, newNarrator=None):
        assert (newNarrator in VISIBLE_ACTORS or newNarrator in INVISIBLE_ACTORS) \
            or newNarrator == None
        if self.sound != None and isinstance(self.sound, str) and len(self.sound) > 0 and not self.soundReady:
            self.sound = switch_to_basename(self.sound)
            self.narrator = newNarrator if newNarrator != None else self.narrator
            self._languages = None

    def export(self):
        data = {}
        if isinstance(self.transcript, MText) and self.transcript.export() != None:
            data["transcript"] = self.transcript.export()
        else:
            return None

        if self.sound != None:
            data["sound"] = self.sound
        if self.narrator != None:
            data["narrator"] = self.narrator
        if isinstance(self.alternative, MText) and self.alternative.export() != None:
            data["alternative"] = self.alternative.export()
        if isinstance(self._languages, list) and len(self._languages) > 0:
            data["languages"] = self._languages
        if self.soundReady:
            data["soundReady"] = self.soundReady

        return data

class Interaction:
    def __init__(self, start="", duration="", actorId=None, content:Content=None, figure=None, \
                 position=None, transform=None, onResult=None, onPoster=None, type=None):
        self.start = start
        self.duration = duration
        self.onResult = onResult
        self.onPoster = onPoster
        self.actorId = actorId
        self.figure = figure
        self.position = position
        self.transform = transform
        self.content = content if isinstance(content, Content) else Content()
        self.type = type

    def copy(self):
        return copy.deepcopy(self)
    
    def merge(self, interaction):
        updated = self.copy()
        if isinstance(interaction, Interaction):
            updated.start = interaction.start if interaction.start != "" else updated.start
            updated.duration = interaction.duration if interaction.duration != "" else updated.duration
            updated.onResult = interaction.onResult if interaction.onResult != None else updated.onResult
            updated.onPoster = interaction.onPoster if interaction.onPoster != None else updated.onPoster
            updated.actorId = interaction.actorId if interaction.actorId != None else updated.actorId
            updated.figure = interaction.figure if interaction.figure != None else updated.figure
            updated.position = interaction.position if interaction.position != None else updated.position
            updated.transform = interaction.transform if interaction.transform != None else updated.transform
            updated.content = interaction.content if interaction.content.export() != None else updated.content
            updated.type = interaction.type if interaction.type != None else updated.type

        return updated
    
    def export(self):
        data = {}
        data["start"] = self.start
        data["duration"] = self.duration
        if isinstance(self.content, Content) and self.content.export()!=None:
            data["content"] = self.content.export()
        if self.onResult != None:
            data["onResult"] = self.onResult
        if self.onPoster != None:
            data["onPoster"] = self.onPoster
        if self.position != None and isinstance(self.position, list):
            data["position"] = self.position
        if self.transform != None and isinstance(self.transform, str) \
            and len(self.transform) > 0:
            data["transform"] = self.transform
        if self.figure != None:
            data["figure"] = self.figure
        if self.actorId != None:
            data["actor"] = self.actorId
        if self.type != None:
            data["type"] = self.type

        if len(data) > 2:
            return data
        else:
            return None

# default interaction with actor posture
class PostureInteraction(Interaction):
    def __init__(self, actorId=-1, figure=None, position=None, transform=None):
        super().__init__(type="motion", actorId=actorId, figure=figure, \
                            position=position, transform=transform)
        
class Event:
    def __init__(self, id=None, scene:Scene=None, board:Board=None, objects:list=None, interactions:list=None):
        self.id = id
        self.scene = scene.copy()
        self.board = board.copy()
        self.objects = [actor.copy() for actor in self.objects if isinstance(actor, Actor)]
        self.interactions = [interaction.copy() for interaction in self.interactions if isinstance(interaction, Interaction)]
    
    def export(self):
        data = {}
        if self.id != None:
            data["id"] = self.id
        if isinstance(self.scene, Scene):
            data["scene"] = self.scene.export()
        if isinstance(self.board, Board):
            data["board"] = self.board.export()
        if isinstance(self.objects, list):
            data["objects"] = [actor.export() for actor in self.objects if isinstance(actor, Actor)]
        if isinstance(self.interactions, list):
            data["interactions"] = [interaction.export() for interaction in self.interactions if isinstance(interaction, Interaction)]
        
        if len(data) > 0:
            return data
        else:
            None

class Actor:
    def __init__(self, name):
        if name in (VISIBLE_ACTORS+INVISIBLE_ACTORS+SCENARIO_ACTORS):
            self.name = name
        elif isinstance(name, str) and name.lower() in ('eily', 'eilly'):
            self.name = 'eilly'
        else:
            self.name = None
    
    @staticmethod
    def load(object):
        if isinstance(object, dict):
            return Actor(object.get("name", None))
        return None

    def match(self, name):
        return self.name!=None \
            and name!=None \
            and (self.name == name \
                or (isinstance(self.name, str) and self.name.lower() in ('eily', 'eilly') \
                and isinstance(name, str) and name.lower() in ('eily', 'eilly')))
    
    def category(self):
        if isinstance(self.name, str):
            if self.name.lower() in VISIBLE_ACTORS:
                return VISIBLE
            elif self.name.lower() in INVISIBLE_ACTORS:
                return INVISIBLE
            elif self.name.lower() in SCENARIO_ACTORS:
                return SCENARIO
        return None

    def export(self):
        data = {}
        if self.name != None:
            data["name"] = self.name
        
        if len(data) > 0:
            return data
        return None


import json
import random
import os
import uuid
import copy

from .utility import *
from .storyprofiles import CHARACTER_FIGURE_ACCESSORY_KEYS, STORY_SCENARIO_STYLES
from .characterpostures import CHARACTER_FIGURES
from .storybase import *

PRODUCTION_ROOT="/story/"
TEST_ROOT="/test/"

class Story:

    def test(self, fileName="testStory.json", localOutputPath=LOCAL_DEFAULT_ROOT, incremental=True):
        for page in self._pages:
            page.exportAudios(
                localOutputPath=localOutputPath, 
                synthesizer=self._synthesizer, 
                cosUploader=self._cosUploader, 
                uploadToCos=True, 
                incremental=incremental
            )

        with open(os.path.join(localOutputPath, fileName), "w") as file:
            json.dump(
                self.export(), file, ensure_ascii=False, indent=4, sort_keys=False
            )
        print(f"Story.test exported to {os.path.join(localOutputPath, fileName)}")
    
    def export(self, debug=False):
        voices = [{"sound": ENDING_SOUND}]
        events = []
        for i, page in enumerate(self._pages):
            if debug:
                print(f"Story.export(), page{i} type {BLUE}{page.pageType}{RESET}")
            pageObject = page.export(voiceOffset=len(voices), pageId=float(len(events)))
            if isinstance(pageObject, dict) \
                and "voices" in pageObject and "events" in pageObject:
                for entry in pageObject["voices"]:
                    entryObject = {}
                    if isinstance(entry, dict):
                        for key in set(entry.keys()) & {"sound", "languages"}:
                            if entry[key] != None:
                                entryObject[key] = entry[key]
                        if len(entryObject) > 0:
                            voices.append(entryObject)
                events = events + pageObject["events"]

        return {"voices": voices, "events": events}

    def exportScripts(self):
        voices = []
        for i, page in enumerate(self._pages):
            pageObject = page.export(voiceOffset=len(voices))
            pageVoices = []
            for voice in pageObject["voices"]:
                if isinstance(voice, dict) \
                    and isinstance(voice.get("transcript", None), (str, dict)) \
                    and len(voice["transcript"]) > 0:
                    pageVoices.append(voice)
            if len(pageVoices) > 0:
                voices = voices + [{"page": i, "voices": pageVoices}]

        return voices

    def exportAudios(self, localOutputPath=LOCAL_DEFAULT_ROOT, uploadToCos=True, incremental=True):
        for page in self._pages:
            page.exportAudios(
                localOutputPath=localOutputPath, 
                synthesizer=self._synthesizer, 
                cosUploader=self._cosUploader, 
                uploadToCos=uploadToCos,
                incremental=incremental
            )

    def exportProduct(self, fileName=None, localOutputPath='./prod'):
        if self._cosUploader == None:
            print("Cos uploader is not available, exit.")
            return

        if not os.path.exists(localOutputPath):
            os.makedirs(localOutputPath)

        storyObject = self.export()
        
        # Copy audios to product path
        for i, voice in enumerate(storyObject["voices"]):
            defaultFileName = voice["sound"]
            storyObject["voices"][i]["sound"] = self._cosUploader.test2product(defaultFileName)
            if isinstance(storyObject["voices"][i].get("languages", None), list) and len(storyObject["voices"][i]["languages"]) > 0:
                for language in storyObject["voices"][i]["languages"]:
                    lingualFileName = defaultFileName[:-3] + language + '.mp3'
                    self._cosUploader.test2product(lingualFileName)


        # Copy images to product path
        for j, event in enumerate(storyObject["events"]):
            if "board" in event and isinstance(event["board"], dict):
                board = event["board"]
                if isinstance(board.get("content", None), dict):
                    if isinstance(board["content"].get("image", None), str) \
                    and len(board["content"]["image"]) > 0:
                        storyObject["events"][j]["board"]["content"]["image"] = self._cosUploader.test2product(board["content"]["image"])
                    elif isinstance(board["content"].get("image", None), dict) \
                    and len(board["content"]["image"]) > 0:
                        for language in board["content"]["image"].keys():
                            storyObject["events"][j]["board"]["content"]["image"][language] = self._cosUploader.test2product(board["content"]["image"][language])
                if isinstance(board.get("contentList", None), list) \
                    and len(board["contentList"]) > 0:
                    for k, contentEntry in enumerate(board["contentList"]):
                        if isinstance(contentEntry.get("image", None), str) \
                            and len(contentEntry["image"]) > 0:
                            storyObject["events"][j]["board"]["contentList"][k]["image"] = self._cosUploader.test2product(contentEntry["image"])
                        elif isinstance(contentEntry.get("image", None), dict) \
                            and len(contentEntry["image"]) > 0:
                            for language in contentEntry["image"].keys():
                                storyObject["events"][j]["board"]["contentList"][k]["image"][language] = self._cosUploader.test2product(contentEntry["image"][language])

        productFileName = fileName if fileName != None else os.path.join(localOutputPath, self.title + ".product.json")
        with open(productFileName, "w") as file:
            json.dump(
                storyObject, file, ensure_ascii=False, indent=4, sort_keys=False
            )
        print(f"Story resource copied from test to production, product story generated as {productFileName}.")

    @staticmethod
    def buildStoryCollection(outputName, storyList):
        storyCollection = {"collection": []}
        for story in storyList:
            storyTitle = story[:len(story)-5] if story.endswith(".json") else story
            storyCollection["collection"].append(storyTitle)
        with open(outputName, "w") as file:
            json.dump(
                storyCollection, file, ensure_ascii=False, indent=4, sort_keys=False
            )
    
    @staticmethod
    def loadFromFile(fileName, locale=DEFAULT_LANGUAGE, **kwargs):
        story = None
        storyId = None
        try:
            with open(fileName, 'r') as f:
                object = json.load(f)
            voices = object["voices"]
            events = object["events"]
            storyId = kwargs.get("storyId", None)
            if len(voices) > 1:
                for i in range(1, len(voices)):
                    folder = voices[i].get("sound", "//").split("/")[-2]
                    if len(folder) == 36: # length of uuid.uuid4()
                        storyId = folder
            storyStyle = None
            validScene = None
            for i in range(len(events)):
                if events[i].get("scene", None) != None and len(events[i]["scene"]) > 0:
                    validScene = events[i]["scene"]
                    break
            for styleKey in STORY_SCENARIO_STYLES.keys():
                for key, value in STORY_SCENARIO_STYLES[styleKey]["scenarios"].items():
                    if value == validScene \
                        or (key == "notes" and value["scene"] == validScene):
                        storyStyle = styleKey
                        break
            defaultNarrator = None
            for event in events:
                if defaultNarrator != None:
                    break
                if isinstance(event.get("objects", None), list):
                    _, _, defaultNarrator, _, _, _ = get_actors(event["objects"])
            kwargs["narrator"] = defaultNarrator if defaultNarrator != None else DEFAULT_NARRATOR
            story = Story(title=os.path.basename(fileName).replace(".json", ""), 
                        storyId=storyId, 
                        style=storyStyle, 
                        locale=locale, 
                        **kwargs)

            pageScenario = "cover"      # 没有样式匹配，设为CoverPage
            for event in events:
                # 获取页面类型
                if "board" in event \
                    and (
                        len(event["board"].get("type", '')) > 0 \
                        or (isinstance(event["board"].get("content", None), dict) \
                            and event["board"]["content"].get("type", None) != None) \
                        or (isinstance(event["board"].get("contentList", None), list) \
                            and any(isinstance(contentEntry, dict) and contentEntry.get('type', None) != None for contentEntry in event["board"]["contentList"])) \
                    ):
                    if isinstance(event["board"].get("contentList", None), list):
                        for entry in event["board"]["contentList"]:
                            if entry.get('type', None) != None:
                                pageScenario = entry['type']
                    pageScenario = event["board"]["type"] \
                        if event["board"].get("type", None) != None \
                        else (event["board"]["content"]["type"] \
                              if event["board"]["content"].get("type", None) != None \
                              else pageScenario)
                else:
                    sceneObject = event.get("scene", None)
                    if "index" in sceneObject and sceneObject["index"] == STORY_SCENARIO_STYLES[storyStyle]["scenarios"]["concentrak"]["index"]:
                        pageScenario = "concentrak"
                    elif isinstance(sceneObject, str) and len(sceneObject) > 0:
                        for key, value in STORY_SCENARIO_STYLES[storyStyle]["scenarios"].items():
                            if isinstance(value, str) and value == sceneObject:
                                pageScenario = key
                    elif "bgColor" in sceneObject and len(sceneObject["bgColor"]) > 0:
                        pageScenario = "blackboard"

                # 创建对应页面
                print(f"Loading page as {pageScenario}")
                # CoverPage
                if pageScenario == "cover":
                    story.createPage(
                        sceneType = pageScenario,
                        source = "",
                        voices = voices,
                        board = event.get("board", None),
                        objects = event.get("objects", None),
                        interactions = event.get("interactions", None)
                        )
                # ClassroomPage
                elif pageScenario == "classroom":
                    story.createPage(
                        sceneType = pageScenario,
                        voices = copy.deepcopy(voices),
                        board = event.get("board", None),
                        objects = event.get("objects", None),
                        interactions = event.get("interactions", None)
                        )
                # BlackboardPage
                elif pageScenario == "blackboard":
                    story.createPage(
                        sceneType = pageScenario,
                        source = "",
                        voices = copy.deepcopy(voices),
                        board = event.get("board", None),
                        objects = event.get("objects", None),
                        interactions = event.get("interactions", None)
                        )

                # ConcentrakPage
                elif pageScenario == "concentrak":
                    story.createPage(
                        sceneType = pageScenario,
                        text = "",
                        voices = copy.deepcopy(voices),
                        board = event.get("board", None),
                        objects = event.get("objects", None),
                        interactions = event.get("interactions", None)
                        )
                # ExamPage
                elif pageScenario == "exam":
                    story.createPage(
                        sceneType = pageScenario,
                        actor = "", 
                        voices = copy.deepcopy(voices),
                        board = event.get("board", None),
                        objects = event.get("objects", None),
                        interactions = event.get("interactions", None)
                    )

                # NotesPage
                elif pageScenario == "notes":
                    story.createPage(
                        sceneType = pageScenario,
                        actor = "", 
                        voices = copy.deepcopy(voices),
                        board = event.get("board", None),
                        objects = event.get("objects", None),
                        interactions = event.get("interactions", None)
                    )

                else:
                    pass            
            
        except Exception as e:
            print("Load story from file exception:\n", e)
            return None
            
        return story

    def __init__(self, title, storyId=None, style="shinkai_makoto", **kwargs):
        self.title = title
        self.storyId = storyId if storyId != None else str(uuid.uuid4())
        self.styles = STORY_SCENARIO_STYLES[style]
        self.locale = kwargs["locale"] if "locale" in kwargs else DEFAULT_LANGUAGE
        self.narrator = kwargs["narrator"] if "narrator" in kwargs else DEFAULT_NARRATOR
        self._pages = []
        self.posterPath = 'test/posters/'
        self.audioPath = 'test/audios/'

        self._cosUploader = kwargs["uploader"] if "uploader" in kwargs else None
        self._synthesizer = kwargs["synthesizer"] if "synthesizer" in kwargs else None

        self._defaultCharacters = CHARACTER_FIGURES

        print(f"Create a new story title: {title}, Id:", self.storyId)

    def setStyle(self, styleName=None, styleObject=None):
        if isinstance(styleName, str) and styleName in STORY_SCENARIO_STYLES and styleName != self.styles:
            newStyles = STORY_SCENARIO_STYLES[styleName]
        elif isinstance(styleObject, dict):
            newStyles = merge_dicts(self.styles, styleObject)
        else:
            print("Not valid styleName or styleObject, return.")
            return

        for page in self._pages:
            page.applyStyle(newStyles)
        
        self.styles = newStyles

    def exportStyle(self):
        return copy.deepcopy(self.styles)

    def getAudioPath(self, fileName):
        return os.path.join("/", self.audioPath, self.storyId, fileName)

    def _getPosturePosition(self, actor, id):
        figureName = self._defaultCharacters[actor][id]
        if "boy" in figureName:
            if "half" in figureName:
                return self.styles["positions"]["right-bottom"]
            elif "standright" in figureName:
                return self.styles["positions"]["right"]
            elif "-stand-" in figureName:
                return self.styles["positions"]["left"]
            else: # head
                return [0, 0]
        elif "girl" in figureName:
            if "half" in figureName:
                return self.styles["positions"]["right-bottom"]
            elif "-stand-" in figureName:
                return self.styles["positions"]["right"]
            else: # head
                return [0, 0]
        else:
            return [0, 0]

        return [0, 0]
    def getUserPostureIdAndPosition(
        self, actor, postures, keyScenario="stand", excludeAccessories=True
    ):
        id = -1
        if type(postures) is int:
            id = postures
        elif type(postures) is list and type(postures[0]) is int:
            id = postures[0]
        elif self._defaultCharacters == None:
            id = 0
        else:
            currentActorFigures = self._defaultCharacters[actor]
            availableFigures = []
            for j, figure in enumerate(currentActorFigures):
                skip = False
                if excludeAccessories:
                    for accessory in CHARACTER_FIGURE_ACCESSORY_KEYS:
                        if accessory in figure:
                            skip = True
                if skip:
                    continue
                if keyScenario in figure and all(keyWord in figure for keyWord in postures):
                    availableFigures.append({"index": j, "figure": figure})
            id = random.choice(availableFigures)["index"] if len(availableFigures) > 0 else 0
        return id, self._getPosturePosition(actor, id)

    def _NewPage(self, sceneType, **kwargs):
        try:
            scene = sceneType.lower()
        except Exception as e:
            print(f"problematic sceneType in type {type(sceneType)}: {sceneType}")
        newPage = None
        if scene == "classroom":
            newPage = ClassroomPage(self, **kwargs)
        elif scene == "blackboard":
            newPage = BlackboardPage(self, **kwargs)
        elif scene == "cover":
            newPage = CoverPage(self, **kwargs)
        elif scene == "exam":
            newPage = ExamPage(self, **kwargs)
        elif scene == "concentrak":
            newPage = ConcentrakPage(self, **kwargs)
        elif scene == "notes":
            newPage = NotesPage(self, **kwargs)
        else:
            print(f"Invalid scenario type {sceneType}, must be one of ('exam', 'notes', 'cover', 'blackboard', 'concentrak', 'classroom')")

        return newPage        

    def createPage(self, sceneType, **kwargs):
        newPage = self._NewPage(sceneType, **kwargs)

        if newPage != None:
            self._pages.append(newPage)

        return newPage

    def createPageAtPos(self, pos, sceneType, **kwargs):
        if pos >= 0 and pos < len(self._pages):
            newPage = self._NewPage(sceneType, **kwargs)
            if newPage != None:
                self._pages.insert(pos, newPage)
        else:
            print("Input pos is out of boundary.")
            newPage = None

        return newPage
    
    def removePageAtPos(self, pos):
        if pos >= 0  and pos < len(self._pages):
            self._pages.pop(pos)

    def getPage(self, pos):
        return self._pages[pos] if (pos >= 0  and pos < len(self._pages)) else None

    def uploadImageToCos(self, source, applyHash=True):
        if self._cosUploader != None:
            if isinstance(source, dict):
                for key in source:
                    source[key] = self._cosUploader.local2cos(source[key], self.storyId, self.posterPath, applyHash)
            else:
                source = self._cosUploader.local2cos(source, self.storyId, self.posterPath, applyHash)
        return source


class Page:
    def __init__(self, pageType, storyInstance):
        self.pageType = pageType
        self.story = storyInstance
        self.narrator = storyInstance.narrator
        self.locale = storyInstance.locale
        self.scene = None
        self.duration = None
        self.transition = None
        self.board = None
        self.objects = []
        self.transcripts = []
        self.mutescripts = []
        self.interactions = []
        self.defaultInteractions = []
        self.actor = None

    def applyStyle(self, styleObject):
        if "scenarios" in styleObject:
            if self.pageType in styleObject["scenarios"]:
                sceneValue = styleObject["scenarios"][self.pageType]
                if isinstance(sceneValue, str):
                    self.scene.value = sceneValue
                elif isinstance(sceneValue, dict):
                    if "scene" in sceneValue:
                        self.scene.value = sceneValue.get("scene", None) if isinstance(sceneValue.get("scene", None), str) else None
                        self.scene.index = sceneValue["scene"].get("index", None) if isinstance(sceneValue.get("scene", None), dict) else None
                        self.scene.bgColor = sceneValue["scene"].get("bgColor", None) if isinstance(sceneValue.get("scene", None), dict) else None
                    else:
                        self.scene.value = None
                        self.scene.index = sceneValue.get("index", None)
                        self.scene.bgColor = sceneValue.get("bgColor", None)

        if "frame" in styleObject:
            self.board.content.border = styleObject["frame"] if self.board.content.border!=None else None
            for i, content in enumerate(self.board.contentList):
                self.board.contentList[i].border = styleObject["frame"] if content.border!=None else None
        if "positions" in styleObject:
            oldPositions = self.story.styles["positions"]
            newPositions = styleObject["positions"]
            if oldPositions != newPositions:
                self.updateActorPosition(oldPositions, newPositions)
        if "popup" in styleObject:
            oldDialogPopup = self.story.styles["popup"]
            newDialogPopup = styleObject["popup"]
            if oldDialogPopup != newDialogPopup:
                self.updateDialogPopup(oldDialogPopup, newDialogPopup)
        if "transform" in styleObject:
            oldTransform = self.story.styles["transform"]
            newTransform = styleObject["transform"]
            if oldTransform != newTransform:
                self.updateActorTransform(oldTransform, newTransform)

    def updateActorPosition(self, oldPositions, newPositions):
        for i, interaction in enumerate(self.defaultInteractions):
            if isinstance(interaction, PostureInteraction):
                currentPositionKey = None
                for key in oldPositions:
                    if oldPositions[key] == interaction.position:
                        currentPositionKey = key
                        break
                if currentPositionKey in newPositions:
                    self.defaultInteractions[i].position = newPositions[currentPositionKey]

    def updateDialogPopup(self, oldDialogPopup, newDialogPopup):
        for i, interaction in enumerate(self.interactions):
            if interaction.content.popup == oldDialogPopup:
                self.interactions[i].content.popup = newDialogPopup
        for i, interaction in enumerate(self.defaultInteractions):
            if interaction.content.popup == oldDialogPopup:
                self.defaultInteractions[i].content.popup = newDialogPopup

    def updateActorTransform(self, oldTransform, newTransform):
        for i, interaction in enumerate(self.defaultInteractions):
            if isinstance(interaction, PostureInteraction) and interaction.transform == oldTransform:
                self.defaultInteractions[i].transform = newTransform

    def _getUserId(self, actor):
        userId = -1
        for i, object in enumerate(self.objects):
            if object["name"].lower() == actor.lower():
                userId = i

        if userId == -1:
            self.objects.append({"name": actor})
            userId = len(self.objects) - 1
        return userId
    
    @staticmethod
    def fit_image_rect(rect, width, height, screenWidth=960.0, screenHeight=540.0):
        # image is wider in ratio
        if width / height > (rect[2] if rect[2] > 1.0 else rect[2]*screenWidth) / (rect[3] if rect[3] > 1.0 else rect[3]*screenHeight):
            height = round((rect[2] if rect[2] > 1.0 else rect[2]*screenWidth) * height / width / (1.0 if rect[3] > 1.0 else screenHeight), 3)
            width = rect[2] * 1.0
        # vice versa, rect is wider in ratio
        else:
            width = round((rect[3] if rect[3] > 1.0 else rect[3]*screenHeight) * width / height / (1.0 if rect[2] > 1.0 else screenWidth), 3)
            height = rect[3] * 1.0

        rect[0] += round(((rect[2] if rect[2] > 1.0 else rect[2]*screenWidth) \
                          - (width if width > 1.0 else width*screenWidth))/screenWidth/2.0, 3)
        rect[1] += round(((rect[3] if rect[3] > 1.0 else rect[3]*screenHeight) \
                          - (height if height > 1.0 else height*screenHeight))/screenHeight/2.0, 3)
        rect[2] = width if width > 1.0 else width * screenWidth
        rect[3] = height if height > 1.0 else height * screenHeight
        
        return rect

    def test(self, fileName="testPage.json", localOutputPath=LOCAL_DEFAULT_ROOT, incremental=True):
        if self.story==None:
            print(f"{RED}No story exists, return.{RESET}")
            return

        self.exportAudios(
            localOutputPath=localOutputPath, 
            synthesizer=self.story._synthesizer, 
            cosUploader=self.story._cosUploader, 
            uploadToCos=True, 
            incremental=incremental
        )

        with open(os.path.join(localOutputPath, fileName), "w") as file:
            json.dump(
                self.export(), file, ensure_ascii=False, indent=4, sort_keys=False
            )
        print(f"Page.test exported to {os.path.join(localOutputPath, fileName)}")

    def __audible_script_list__(self):
        return []

    def exportAudios(self, localOutputPath=LOCAL_DEFAULT_ROOT, synthesizer=None, cosUploader=None, uploadToCos=True, incremental=True):
        if self.story==None:
            print(f"{RED}No story exists, return.{RESET}")
            return
        directory = os.path.join(localOutputPath, self.story.storyId)
        if not os.path.exists(directory):
            os.makedirs(directory)

        currentSynthesizer = synthesizer if synthesizer!=None else self.story._synthesizer if self.story!=None else None
        currentCosUploader = cosUploader if cosUploader!=None else self.story._cosUploader if self.story!=None else None
        if currentCosUploader == None:
            print(f"{YELLOW}No COS uploader available, ignore uploading.{RESET}")

        if currentSynthesizer:
            for transcripts in self.__audible_script_list__():
                if len(transcripts) < 1:
                    continue
                for i, scriptObject in enumerate(transcripts):
                    if isinstance(scriptObject, Script):
                        if isinstance(scriptObject.sound, str) \
                            and (os.path.basename(scriptObject.sound) == scriptObject.sound or not incremental) \
                            and (not scriptObject.soundReady):
                            try:
                                fileName = os.path.basename(scriptObject.sound)
                                character = scriptObject.narrator
                                transcriptObject = MText(scriptObject.transcript).export()
                                alternativeObject = MText(scriptObject.alternative).export()
                                if isinstance(transcriptObject, str):
                                    transcriptStr = alternativeObject[DEFAULT_LANGUAGE] \
                                        if isinstance(alternativeObject, dict) \
                                            and len(alternativeObject.get(DEFAULT_LANGUAGE, '')) > 0 \
                                        else (
                                            alternativeObject \
                                                if isinstance(alternativeObject, str) \
                                                    and len(alternativeObject) > 0 \
                                                else transcriptObject
                                        )
                                    currentSynthesizer.synthesizeFile(
                                        character, remove_emojis(transcriptStr), DEFAULT_LANGUAGE, directory, fileName
                                    )
                                    localOutputFileName = os.path.join(directory, fileName)

                                    if currentCosUploader != None and uploadToCos:
                                        currentCosUploader.local2cos(localOutputFileName, self.story.storyId, self.story.audioPath)
                                        transcripts[i].sound = self.story.getAudioPath(fileName)
                                    transcripts[i]._languages = None
                                elif isinstance(transcriptObject, dict):
                                    processedLanguages = []
                                    for language in transcriptObject:
                                        transcriptStr = alternativeObject[language] \
                                            if isinstance(alternativeObject, dict) \
                                                and len(alternativeObject.get(language, '')) > 0 \
                                            else (
                                                alternativeObject \
                                                    if language == DEFAULT_LANGUAGE \
                                                        and isinstance(alternativeObject, str) \
                                                        and len(alternativeObject) > 0 \
                                                    else transcriptObject[language]
                                            )
                                        if language != DEFAULT_LANGUAGE:
                                            fileName = fileName[:-3]+language+".mp3"                                            
                                        currentSynthesizer.synthesizeFile(
                                            character, remove_emojis(transcriptStr), language, directory, fileName
                                        )
                                        localOutputFileName = os.path.join(directory, fileName)

                                        if currentCosUploader != None and uploadToCos:
                                            currentCosUploader.local2cos(localOutputFileName, self.story.storyId, self.story.audioPath)
                                            if language == DEFAULT_LANGUAGE:
                                                transcripts[i].sound = self.story.getAudioPath(fileName)
                                            else:
                                                processedLanguages.append(language)
                                    transcripts[i]._languages = processedLanguages
                            except Exception as e:
                                print(f"Synthesize & upload script failed for [{scriptObject.transcript}]\n", e)
                                continue
        else:
            print(f"{RED}No synthesizer available, return.{RESET}")

    def export(self, voiceOffset, pageId):
        raise NotImplementedError("Subclasses must implement export()")

    def exportScripts(self):
        return self.export()["voices"]

##### 问答页面 #####
class ExamPage(Page):
    # onResult: -2
    class InitInteraction(Interaction):
        ON_RESULT = -2
        POPUP = 4
        def __init__(self, actorId=None, text=None, voice=None):
            super().__init__(onResult=self.ON_RESULT, actorId=actorId, type="talk", 
                                content=Content(
                                    popup=self.POPUP, 
                                    voice=voice, 
                                    text=text
                                )
                            )

    # onResult: -1
    class ErrorInteraction(Interaction):
        ON_RESULT = -1
        POPUP = 4
        def __init__(self, actorId=None, text=None, voice=None):
            super().__init__(onResult=self.ON_RESULT, actorId=actorId, type="talk",
                                content=Content(
                                    popup=self.POPUP, 
                                    voice=voice, 
                                    text=text
                                )
                            )

    # popup = 2, onResult: self.correctAnswerId
    class SuccessInteraction(Interaction):
        POPUP = 2
        def __init__(self, actorId=None, onResult=None, text=None, voice=None):
            super().__init__(onResult=onResult, actorId=actorId, type="talk", 
                                content=Content(
                                    popup=self.POPUP, 
                                    voice=voice, 
                                    text=text
                                )
                            )
    
    def __init__(self, storyInstance, actor=None, postures=["smilesay"], sound=None, **kwargs):
        super().__init__("exam", storyInstance)
        self.scene = Scene.load(self.story.styles["scenarios"][self.pageType])
        self.board = Board(type=self.pageType)
        self.defaultObject = self.pageType
        self.questionTranscript = None
        self.questionInteractions = []
        self.boardContentListScripts = []

        if all(key in kwargs for key in ("voices", "board", "objects", "interactions")):
            voices = kwargs["voices"]
            self.scene = Scene.load(kwargs["scene"]) if kwargs.get("scene", None) != None else self.scene
            self.board = Board.load(kwargs["board"])
            if self.board.content.options != None:
                count = len(MList(self.board.content.options).firstValidPair()[1])
                for i in range(count):
                    self.mutescripts.append(Script(transcript=self.board.content.options.getMTextByPos(i)))
            self.objects = kwargs["objects"]
            self.actor, actorId, _, _, _, _ = get_actors(self.objects)
            for interaction in kwargs["interactions"]:
                if isinstance(interaction, dict):
                    if ("figure" in interaction and interaction.get("figure", -1) > -1) and "position" in interaction:
                        self.defaultInteractions.append(PostureInteraction(
                            actorId = interaction["actor"] if interaction.get("actor", -1) > -1 else actorId,
                            position = interaction["position"],
                            figure = interaction["figure"],
                            transform = interaction["transform"] \
                                if (isinstance(interaction.get("transform", None), str) \
                                    and len(interaction["transform"]) > 0) \
                                else None
                        ))

                    # onResult in (-2, 0), initInteraction
                    if "content" in interaction and \
                        "onResult" in interaction and (interaction["onResult"][0] if isinstance(interaction["onResult"], list) else interaction["onResult"]) in (-2, 0):
                        content = interaction["content"]
                        if isinstance(content, dict) and "voice" in content:
                            voiceId = content["voice"]
                            self.soundFile = voices[voiceId]["sound"]
                            self.questionTranscript = Script(
                                sound = self.soundFile,
                                transcript = MText(self.board.content.question),
                                narrator = self.actor if self.actor!=None else self.narrator,
                                languages = voices[voiceId]["languages"] if "languages" in voices[voiceId] else None)
                            self.defaultInteractions.append(ExamPage.InitInteraction(
                                actorId = self._getUserId(self.defaultObject),
                                text = MText(self.board.content.question) if self.board.content.question!=None else None,
                                voice = 0
                            ))
                    elif "onResult" in interaction:
                        onResultValue = (interaction["onResult"][0] if isinstance(interaction["onResult"], list) else interaction["onResult"])
                        textValue = interaction["content"].get("text", None) \
                            if ("content" in interaction and isinstance(interaction["content"], dict)) \
                            else None
                        
                        # onResultValue > 0, successInteraction
                        if onResultValue > 0:
                            self.correctAnswerId = onResultValue
                            if "content" in interaction and isinstance(interaction["content"].get("popup", None), int):
                                popup = interaction["content"]["popup"]
                                if popup == 2:
                                    self.defaultInteractions.append(ExamPage.SuccessInteraction(
                                        actorId = self._getUserId(self.defaultObject),
                                        onResult = onResultValue,
                                        text = MText(textValue) if textValue!=None else None
                                    ))
                                if popup == 4:
                                    self.defaultInteractions.append(Interaction(
                                        actorId = self._getUserId(self.defaultObject),
                                        onResult = onResultValue,
                                        content = Content(
                                            text = MText(textValue) if textValue!=None else None,
                                            popup = popup
                                        ),
                                        type = "talk"
                                    ))
                        # onResultValue == -1, errorInteraction
                        elif onResultValue == -1:
                            self.defaultInteractions.append(ExamPage.ErrorInteraction(
                                actorId = self._getUserId(self.defaultObject),
                                text = MText(textValue) if textValue!=None else None
                            ))
                        if textValue!=None and len(textValue) > 0:
                            self.mutescripts.append(Script(transcript=MText(textValue)))
                    elif interaction.get("type", None) in ("motion", "talk") and "figure" in interaction:
                        postures = interaction["figure"]
                        if isinstance(interaction.get("content", None), dict) and "voice" in interaction["content"]:
                            content = interaction["content"]
                            voiceId = content["voice"]
                            if MText(content.get("text", "")) == self.board.content.question and voiceId > 0:
                                self.soundFile = voices[voiceId]["sound"]
                                self.questionTranscript = Script(
                                    sound = self.soundFile,
                                    transcript = self.board.content.question.copy(),
                                    narrator = self.actor if self.actor!=None else self.narrator,
                                    languages = voices[voiceId].get("languages", None)
                                )
                                                                
                                self.questionInteractions.append(ExamPage.InitInteraction(
                                    actorId = self._getUserId(self.defaultObject),
                                    text = self.board.content.question,
                                    voice = voiceId
                                ))
        else:
            if actor != None:
                self.setActor(actor, postures, **kwargs)
            self.soundFile = f"voice-{uuid.uuid4()}.mp3" if sound == None else sound
            
    def setActor(self, actor, postures=["smilesay"], **kwargs):
        if actor in VISIBLE_ACTORS:
            if self.actor != actor:
                update_visible_actor(self.objects, actor)
                if isinstance(self.questionTranscript, Script):
                    self.questionTranscript.reset2basename(actor)
                self.actor = actor

            posture_list = (
                [postures]
                if isinstance(postures, (str, int))
                else [postures[0]] \
                    if (isinstance(postures, list) and isinstance(postures[0], int))
                    else postures
            )
            figure, position = self.story.getUserPostureIdAndPosition(actor, posture_list, keyScenario="half")
            if isinstance(kwargs.get("position", None), list) and len(kwargs["position"]) == 2:
                position = kwargs["position"]
            scale = None
            if isinstance(kwargs.get("scale", None), (int, float)) and kwargs["scale"] > 0:
                scale = f"scale({kwargs['scale']})"
            hasPosture = False
            for i, interaction in enumerate(self.defaultInteractions):
                if isinstance(interaction, PostureInteraction):
                    self.defaultInteractions[i].actorId = self._getUserId(self.actor)
                    self.defaultInteractions[i].position = position
                    self.defaultInteractions[i].transform = scale if scale != None else interaction.transform
                    self.defaultInteractions[i].figure = figure
                    hasPosture = True
            if not hasPosture:
                self.defaultInteractions.append(PostureInteraction(
                    actorId = self._getUserId(self.actor),
                    position = position,
                    transform = scale if scale != None else "scale(1.5)",
                    figure = figure
                ))
        elif actor==None and self.actor!=None:
            for i, interaction in enumerate(self.defaultInteractions):
                if isinstance(interaction, PostureInteraction):
                    self.defaultInteractions.pop(i)
            if isinstance(self.questionTranscript, Script):
                self.questionTranscript.reset2basename(self.narrator)
            self.actor = None

    def setQuestion(self, question:MText, options:MList=[], answers=None, **kwargs):
        newQuestion = MText(question)
        newOptions = MList(options, DEFAULT_LANGUAGE)
        newAnswers = MList(answers, DEFAULT_LANGUAGE)

        hasChinese = False
        for i in range(len(newOptions.getListByLanguage(DEFAULT_LANGUAGE))):
            if has_chinese_char(newOptions.getListByLanguage(DEFAULT_LANGUAGE)[i]):
                hasChinese = True
            self.mutescripts.append(Script(transcript=newOptions.getMTextByPos(i)))

        if not hasChinese:
            newAnswers = MList(newAnswers.getListByLanguage(DEFAULT_LANGUAGE))
            newOptions = MList(newOptions.getListByLanguage(DEFAULT_LANGUAGE))

        fontSize = kwargs.get("fontSize", "20px")
        if isinstance(fontSize, int):
            fontSize = str(fontSize) + "px"

        self.board.content.fontSize=fontSize
        self.board.content.fontColor=kwargs.get("fontColor", "white")
        self.board.content.question=newQuestion if question != None else self.board.content.question
        self.board.content.options=newOptions
        self.board.content.answer=newAnswers
        self.board.content.colsPerRow=kwargs.get("colsPerRow", 1)
        self.board.rect=kwargs.get("rect", [0, 0, 1, 1])

        self.correctAnswerId = 0
        for i, option in enumerate(newOptions.getListByLanguage(DEFAULT_LANGUAGE)):
            for entry in (newAnswers.getListByLanguage(DEFAULT_LANGUAGE)):
                if entry == option:
                    self.correctAnswerId += 2**i

        questionTranscript = kwargs.get("alternativeText", None) if kwargs.get("alternativeText", None)!=None else newQuestion
        # 如果新问题文字与原问题文字不同。则需先生成test语音，下一步再更新production上语音
        if self.questionTranscript == None:
            self.questionTranscript = Script(
                sound = f"voice-{uuid.uuid4()}.mp3",
                transcript = questionTranscript,
                narrator = self.actor if self.actor!=None else self.narrator
            )
        elif self.questionTranscript.transcript != questionTranscript:
            self.questionTranscript.transcript = questionTranscript
            self.questionTranscript.sound = f"voice-{uuid.uuid4()}.mp3" \
                if self.questionTranscript.sound == None or len(self.questionTranscript.sound) == 0 \
                else self.questionTranscript.sound
            self.questionTranscript.reset2basename(self.actor if self.actor!=None else self.narrator)

        self.questionInteractions = []

        # 初始化页面行为 onResult: -2
        self.questionInteractions.append(ExamPage.InitInteraction(
            actorId = self._getUserId(self.defaultObject),
            text = self.board.content.question,
            voice = 0
        ))

        if isinstance(newOptions.firstValidPair()[1], list) and len(newOptions.firstValidPair()[1]) > 0:
            # 错误答案提示行为 onResult: -1
            self.questionInteractions.append(ExamPage.ErrorInteraction(
                actorId = self._getUserId(self.defaultObject),
                text = MText(kwargs.get("alwaysTruePrompt", kwargs.get("onErrorPrompt", {DEFAULT_LANGUAGE: "再想想"})))
            ))

            # 正确答案行为 onResult: 由所有正确答案id计算所得
            if self.correctAnswerId > 0:
                self.questionInteractions.append(ExamPage.SuccessInteraction(
                    onResult=self.correctAnswerId,
                    actorId=self._getUserId(self.defaultObject)
                ))
                if kwargs.get("onSuccessPrompt", None) != None:
                    self.defaultInteractions.append(Interaction(
                                            actorId = self._getUserId(self.defaultObject),
                                            onResult = self.correctAnswerId,
                                            content = Content(
                                                text = MText(kwargs["onSuccessPrompt"]),
                                                popup = 4
                                            ),
                                            type = "talk"
                                        ))
                    self.mutescripts.append(Script(transcript=MText(kwargs["onSuccessPrompt"])))

    def setFontSize(self, size):
        self.board.content.fontSize = str(size)+"px" if isinstance(size, int) else size

    def setColsPerRow(self, colsPerRow):
        self.board.content.colsPerRow = colsPerRow

    def setRect(self, rect):
        self.board.rect = rect

    def setFontColor(self, color):
        self.board.content.fontColor = color

    def addImage(self, source:Union[str, dict], rect:list=[0, 0, 1, 1], autoFit:bool=True, uploadToCos:bool=True, **kwargs):
        assert isinstance(rect, list) and len(rect) >= 4
        validUrl = False
        if is_valid_http_url(MText(source).firstValidPair()[1]):
            width = rect[2]
            height = rect[3]
            validUrl = True
        else:
            width, height = get_image_size(MText(source).firstValidPair()[1])
        assert width > 0 and height > 0
        mSource = MText(source) if source!=None else None
        mCaption = MText(kwargs.get("caption", "")) # Need value for caption to display image in contentList

        if not validUrl:
            if autoFit:
                rect = Page.fit_image_rect(rect, width, height)

            if uploadToCos:
                if isinstance(mSource.data, dict):
                    for key in mSource.data:
                        mSource.data[key] = self.story.uploadImageToCos(mSource.data[key], applyHash=kwargs.get("applyHash", True))
                else:
                    mSource.data = self.story.uploadImageToCos(mSource.data, applyHash=kwargs.get("applyHash", True))

        self.board.contentList.append(Content(
            image=mSource,
            rect=rect,
            caption=mCaption
        ))

        self.boardContentListScripts.append(Script(transcript=mCaption))

    def updateImage(self, pos:int, source:Union[str, dict], rect:list=[0, 0, 1, 1], autoFit:bool=True, uploadToCos:bool=True, **kwargs):
        if pos < len(self.board.contentList) and pos >= 0:
            assert isinstance(rect, list) and len(rect) >= 4
            validUrl = False
            if is_valid_http_url(MText(source).firstValidPair()[1]):
                width = rect[2]
                height = rect[3]
                validUrl = True
            else:
                width, height = get_image_size(MText(source).firstValidPair()[1])
            assert width > 0 and height > 0
            mSource = MText(source) if source!=None else None
            mCaption = MText(kwargs.get("caption", \
                                        self.board.contentList[pos].caption \
                                            if self.board.contentList[pos].caption != None \
                                            else ""
                                        )
                            )
            if isinstance(self.board.contentList[pos].caption, MText):
                mCaption = self.board.contentList[pos].caption.merge(mCaption)

            if not validUrl:
                if autoFit:
                    rect = Page.fit_image_rect(rect, width, height)

                if uploadToCos:
                    if isinstance(mSource.data, dict):
                        for key in mSource.data:
                            mSource.data[key] = self.story.uploadImageToCos(mSource.data[key], applyHash=kwargs.get("applyHash", True))
                    else:
                        mSource.data = self.story.uploadImageToCos(mSource.data, applyHash=kwargs.get("applyHash", True))

            self.board.contentList[pos].image = self.board.contentList[pos].image.merge(mSource)
            self.board.contentList[pos].rect = rect
            self.board.contentList[pos].caption = mCaption
            
            self.boardContentListScripts[pos] = Script(transcript=mCaption)

    def removeImage(self, pos):
        if pos < len(self.board.contentList) and pos >= 0:
            self.board.contentList.pop(pos)
            self.boardContentListScripts.pop(pos)

    def __audible_script_list__(self):
        return [[self.questionTranscript]]

    def export(self, voiceOffset=0, pageId=0.0):
        outTranscripts = [transcript.export() for transcript in ([self.questionTranscript] + self.boardContentListScripts + self.mutescripts) \
                         if (isinstance(transcript, Script) and transcript.export()!=None)]

        outInteractions = []
        for interaction in self.defaultInteractions + self.questionInteractions:
            newInteraction = None
            if interaction.content.voice != None and interaction.content.voice > -1:
                newInteraction = interaction.copy()
                newInteraction.content.voice += voiceOffset
            exported = newInteraction.export() if newInteraction != None else interaction.export() 
            if exported != None:
                outInteractions.append(exported)

        eventDict = {
            "id": pageId,
            "scene": self.scene.export(),
            "board": self.board.export(),
            "objects": copy.deepcopy(self.objects),
            "interactions": outInteractions
        }

        if self.duration is not None:
            eventDict["duration"] = self.duration

        if self.transition is not None:
            eventDict["transition"] = self.transition

        return {
            "voices": outTranscripts,
            "events": [eventDict]
        }

##### 总结页面 #####
class NotesPage(Page):
    def __init__(self, storyInstance, actor=None, postures=["smilesay"], endingEffect=True, sound=None, **kwargs):
        super().__init__("notes", storyInstance)
        self.scene = Scene.load(self.story.styles["scenarios"][self.pageType]["scene"])
        self.board = Board(
            content = Content(html=MHTML.loadFromText(self.story.styles["scenarios"][self.pageType]["htmlTemplate"])),
            rect = self.story.styles["scenarios"][self.pageType]["rect"],
            type = self.pageType
        )
        self.defaultObject = self.pageType

        if all(key in kwargs for key in ("voices", "board", "objects", "interactions")):
            voices = kwargs["voices"]
            self.scene = Scene.load(kwargs["scene"]) if kwargs.get("scene", None) != None else self.scene
            self.board = Board.load(kwargs["board"])
            self.objects = kwargs["objects"]
            self.actor, actorId, self.narrator, _, defaultObject, _ = get_actors(self.objects)
            self.defaultObject = defaultObject if defaultObject != None else self.narrator if self.narrator != None else self.defaultObject
            self.soundFile = f"voice-{uuid.uuid4()}.mp3" if sound == None else sound
            self.soundFileLanguages = None
            self.endingEffect = False
            postures = 0
            transform = None

            for interaction in kwargs["interactions"]:
                if isinstance(interaction, dict):
                    if isinstance(interaction.get("content", None), dict) and "voice" in interaction["content"]:
                        value = interaction["content"]["voice"]
                        if value == 0:
                            self.endingEffect = True
                        elif value > 0:
                            self.soundFile = voices[value]["sound"]
                            self.soundFileLanguages = voices[value].get("languages", None)
                    if "type" in interaction and "figure" in interaction:
                        value = interaction["figure"]
                        if interaction["type"] == "motion":
                            postures = value
                        elif interaction["type"] == "talk":
                            if interaction["actor"] == actorId and value > -1: 
                                postures = value
                        if "transform" in interaction:
                            transform = interaction["transform"]
            figure, position = self.story.getUserPostureIdAndPosition(self.actor, postures, keyScenario="half")
            self.defaultInteractions.append(
                PostureInteraction(
                    position = position,
                    transform = transform if transform != None else "scale(1.5)",
                    figure = figure,
                    actorId = actorId
                ))
            
        else:
            self.setActor(actor, postures, **kwargs)
            self.endingEffect = endingEffect
            self.soundFile = f"voice-{uuid.uuid4()}.mp3" if sound == None else sound
            self.soundFileLanguages = kwargs["languages"] if isinstance(kwargs.get("languages", None), list) else None

        self.endingInteraction = Interaction(
                duration = "auto",
                content = Content(
                    voice = 0,
                    text = MText("")
                ),
                actorId = self._getUserId(self.defaultObject),
                type = "talk"
            )

    def setActor(self, actor, postures=["smilesay"], **kwargs):
        if actor in VISIBLE_ACTORS:
            if self.actor != actor:
                update_visible_actor(self.objects, actor)
                [script.reset2basename(actor) for script in self.transcripts if isinstance(script, Script)]
                self.actor = actor

            posture_list = (
                [postures]
                if isinstance(postures, (str, int))
                else [postures[0]] \
                    if (isinstance(postures, list) and isinstance(postures[0], int))
                    else postures
            )
            figure, position = self.story.getUserPostureIdAndPosition(actor, posture_list, keyScenario="half")
            if isinstance(kwargs.get("position", None), list) and len(kwargs["position"]) == 2:
                position = kwargs["position"]
            scale = None
            if isinstance(kwargs.get("scale", None), (int, float)) and kwargs["scale"] > 0:
                scale = f"scale({kwargs['scale']})"
            hasPosture = False
            for i, interaction in enumerate(self.defaultInteractions):
                if isinstance(interaction, PostureInteraction):
                    self.defaultInteractions[i].actorId = self._getUserId(self.actor)
                    self.defaultInteractions[i].position = position
                    self.defaultInteractions[i].transform = scale if scale != None else interaction.transform
                    self.defaultInteractions[i].figure = figure
                    hasPosture = True
            if not hasPosture:
                self.defaultInteractions.append(PostureInteraction(
                    actorId = self._getUserId(self.actor),
                    position = position,
                    transform = scale if scale != None else "scale(1.5)",
                    figure = figure
                ))
        elif actor==None and self.actor!=None:
            for i, interaction in enumerate(self.defaultInteractions):
                if isinstance(interaction, PostureInteraction):
                    self.defaultInteractions.pop(i)
            self.actor = None

    def addBullet(self, text):
        if isinstance(self.board.content.html, MHTML):
            self.board.content.html.addBullet(text)
            self.soundFileLanguages = None
            if isinstance(self.soundFile, str):
                self.soundFile = switch_to_basename(self.soundFile)

    def updateBullet(self, pos, text):
        if isinstance(self.board.content.html, MHTML):
            if self.board.content.html.updateBullet(pos, text):
                self.soundFileLanguages = None
                if isinstance(self.soundFile, str):
                    self.soundFile = switch_to_basename(self.soundFile)

    def removeBullet(self, pos):
        if isinstance(self.board.content.html, MHTML):
            if self.board.content.html.removeBullet(pos):
                self.soundFileLanguages = None
                if isinstance(self.soundFile, str):
                    self.soundFile = switch_to_basename(self.soundFile)

    def setEndingEffect(self, on: bool):
        self.endingEffect = on

    def exportAudios(self, localOutputPath=LOCAL_DEFAULT_ROOT, synthesizer=None, cosUploader=None, uploadToCos=True, incremental=True):
        if self.story==None:
            print(f"{RED}No story exists, return.{RESET}")
            return
        directory = os.path.join(localOutputPath, self.story.storyId)
        if not os.path.exists(directory):
            os.makedirs(directory)

        currentSynthesizer = synthesizer if synthesizer!=None else self.story._synthesizer if self.story!=None else None
        currentCosUploader = cosUploader if cosUploader!=None else self.story._cosUploader if self.story!=None else None
        if currentCosUploader == None:
            print(f"{YELLOW}No COS uploader available, ignore uploading.{RESET}")

        if currentSynthesizer:
            if isinstance(self.soundFile, str) \
                and (os.path.basename(self.soundFile) == self.soundFile or not incremental):
                try:
                    fileName = os.path.basename(self.soundFile)
                    character = self.actor if self.actor!=None else self.narrator
                    transcriptObject = MText(self.board.content.html.exportScripts()).export()
                    if isinstance(transcriptObject, str):
                        transcriptStr = transcriptObject
                        currentSynthesizer.synthesizeFile(
                            character, remove_emojis(transcriptStr), DEFAULT_LANGUAGE, directory, fileName
                        )
                        localOutputFileName = os.path.join(directory, fileName)

                        if currentCosUploader != None and uploadToCos:
                            currentCosUploader.local2cos(localOutputFileName, self.story.storyId, self.story.audioPath)
                            self.soundFile = self.story.getAudioPath(fileName)
                        self.soundFileLanguages = None
                    elif isinstance(transcriptObject, dict):
                        transcriptDict = transcriptObject
                        processedLanguages = []
                        for language in transcriptDict:
                            transcriptStr = transcriptDict[language]
                            currentSynthesizer.synthesizeFile(
                                character, remove_emojis(transcriptStr), language, directory, fileName
                            )
                            localOutputFileName = os.path.join(directory, fileName)

                            if currentCosUploader != None and uploadToCos:
                                currentCosUploader.local2cos(localOutputFileName, self.story.storyId, self.story.audioPath)
                                self.soundFile = self.story.getAudioPath(fileName)
                                if language!=DEFAULT_LANGUAGE:
                                    processedLanguages.append(language)
                        self.soundFileLanguages = processedLanguages
                except Exception as e:
                    print(f"Synthesize & upload script failed for [{self.board.content.html.exportScripts()}]\n", e)
        else:
            print(f"{RED}No synthesizer available, return.{RESET}")

    def export(self, voiceOffset=0, pageId=0.0):
        outTranscripts = []
        if isinstance(self.board.content, Content) and self.board.content.export()!=None:
            contentScript = Script(
                sound=self.soundFile,
                transcript=MText(self.board.content.html.exportScripts()),
                narrator=self.actor if self.actor!=None else self.narrator,
                languages = self.soundFileLanguages if isinstance(self.soundFileLanguages, list) else None
                )
            outTranscripts.append(contentScript.export())
        if isinstance(self.board.contentList, list) and len(self.board.contentList) > 0:
            for i, entry in enumerate(self.board.contentList):
                entryScript = Script(
                    sound=self.soundFile if i==0 else None,
                    transcript=MText(entry.html.exportScripts()),
                    narrator=(self.actor if self.actor!=None else self.narrator) if i==0 else None,
                    languages = (self.soundFileLanguages if isinstance(self.soundFileLanguages, list) else None) if i==0 else None
                    )
                outTranscripts.append(entryScript.export())

        notesInteraction = Interaction(
                content = Content(
                    voice = 0 + voiceOffset
                ),
                type = "talk",
                actorId = self._getUserId(self.defaultObject)
            )
        outInteractions = [interaction.export() for interaction in (self.defaultInteractions + [notesInteraction] + ([self.endingInteraction] if self.endingEffect else []))]

        eventDict = {
            "id": pageId,
            "scene": self.scene.export(),
            "board": self.board.export(),
            "objects": copy.deepcopy(self.objects),
            "interactions": outInteractions
        }

        if self.duration is not None:
            eventDict["duration"] = self.duration

        if self.transition is not None:
            eventDict["transition"] = self.transition

        return {
            "voices": outTranscripts,
            "events": [eventDict]
        }


##### 概念页面 #####
class ConcentrakPage(Page):
    class ConcentrakInteraction(Interaction):
        def __init__(self, actorId=-1, text:MText=None):
            super().__init__(actorId=actorId, content=Content(popup=6, text=text))

    def __init__(self, storyInstance, **kwargs):
        super().__init__("concentrak", storyInstance)
        self.scene = Scene.load(self.story.styles["scenarios"][self.pageType])
        self.defaultObject = self.pageType

        if all(key in kwargs for key in ("voices", "board", "objects", "interactions")):
            voices = kwargs["voices"]
            self.objects = kwargs["objects"]
            self.scene = Scene.load(kwargs["scene"]) if kwargs.get("scene", None) != None else self.scene
            self.board = Board.load(kwargs["board"])
            for interaction in kwargs["interactions"]:
                if isinstance(interaction, dict):
                    if "content" in interaction:
                        content = interaction["content"]
                        if isinstance(content, dict):
                            if "popup" in content and "text" in content:
                                text = content["text"]
                                if content["popup"] == 6:
                                    self.setTitle(text)
                                elif content["popup"] == 4 and "voice" in content and content["voice"] > 0:
                                    voiceId = content["voice"]
                                    if "actor" in interaction:
                                        actorId = interaction["actor"]
                                        self.transcripts.append(Script(
                                            sound = voices[voiceId]["sound"],
                                            transcript = MText(text),
                                            narrator = self.objects[actorId]["name"],
                                            languages = voices[voiceId]["languages"] \
                                                if isinstance(voices[voiceId].get("languages", None), list) \
                                                else None 
                                            ))
                                        self.interactions.append(Interaction(
                                            duration = "auto",
                                            content = Content(
                                                popup = 4,
                                                text = MText(text),
                                                voice = len(self.transcripts) - 1
                                            ),
                                            actorId = actorId,
                                            type = "talk"))
        else:
            self.board = Board()
            if kwargs.get("title", None) != None:
                self.setTitle(kwargs["title"])
    
    def setTitle(self, text):
        if len(self.mutescripts) == 0:
            self.mutescripts.append(Script(transcript=MText(text)))
            self.defaultInteractions.append(ConcentrakPage.ConcentrakInteraction(
                    actorId = self._getUserId(self.defaultObject),
                    text = MText(text)
                ))
        else:
            self.mutescripts[0].transcript=MText(text)
            self.defaultInteractions[0].content.text = MText(text)

    def addNarration(self, text, narrator=None, alternativeText=None, sound=None, **kwargs):
        narrator = narrator if narrator != None else self.narrator if self.narrator != None else self.actor
        self.transcripts.append(Script(
            sound = f"voice-{uuid.uuid4()}.mp3" if sound == None else sound,
            transcript = MText(text),
            narrator = narrator,
            alternative = MText(alternativeText) if alternativeText!= None else None,
            languages = kwargs["languages"] if isinstance(kwargs.get("languages", None), list) else None,
            soundReady=bool(kwargs.get("soundReady", False))
            ))          
        
        self.interactions.append(Interaction(
            duration = "auto",
            content = Content(
                popup = 4,
                text = MText(text),
                voice = len(self.transcripts) - 1
            ),
            figure = -1 if narrator in VISIBLE_ACTORS else None,
            actorId = self._getUserId(narrator),
            type = "talk"
        ))

    def updateNarration(self, pos, text=None, narrator=None, alternativeText=None):
        transcript = text if isinstance(text, dict) else ({self.locale: text} if text != None else None)
        alternative = alternativeText if isinstance(alternativeText, dict) else ({self.locale: alternativeText} if alternativeText != None else None)
        if pos < len(self.transcripts) and pos >= 0 and (transcript!=None or narrator!=None or alternative!=None):
            self.transcripts[pos].reset2basename()
            if transcript != None:
                self.transcripts[pos].transcript = MText(self.transcripts[pos].transcript).merge(text)
                self.interactions[pos].content.text = MText(self.interactions[pos].content.text).merge(text)
            if alternative != None:
                self.transcripts[pos].alternative = MText(self.transcripts[pos].alternative).merge(text)
            if narrator != None:
                self.transcripts[pos].reset2basename(narrator)
                self.interactions[pos].actorId = self._getUserId(narrator)
                self.interactions[pos].figure = -1 if narrator in VISIBLE_ACTORS else None

    def removeNarration(self, pos):
        if pos < len(self.transcripts) and pos >= 0:
            self.transcripts.pop(pos)
            self.interactions.pop(pos)
            if pos < len(self.interactions):
                for i, interaction in enumerate(self.interactions[pos:]):
                    self.interactions[i].content.voice = interaction.content.voice - 1

    def __audible_script_list__(self):
        return [self.transcripts]

    def export(self, voiceOffset=0, pageId=0.0):
        outTranscripts = [transcript.export() \
                            for transcript in self.transcripts + self.mutescripts \
                            if transcript is not None and transcript.export() is not None]
        
        outInteractions = []
        for interaction in (self.defaultInteractions + self.interactions):
            newInteration = None
            if interaction.content.voice != None and interaction.content.voice > -1:
                newInteration = interaction.copy()
                newInteration.content.voice += voiceOffset
            exported = newInteration.export() if newInteration != None else interaction.export() 
            if exported != None:
                outInteractions.append(exported)

        eventDict = {
            "id": pageId,
            "scene": self.scene.export(),
            "board": self.board.export(),
            "objects": copy.deepcopy(self.objects),
            "interactions": outInteractions
        }

        if self.duration is not None:
            eventDict["duration"] = self.duration

        if self.transition is not None:
            eventDict["transition"] = self.transition

        return {
            "voices": outTranscripts,
            "events": [eventDict]
        }


##### 黑板页面 #####
class BoardPage(Page):
    def __init__(self, pageType, storyInstance, **kwargs):
        assert pageType in ("cover", "blackboard", "classroom")
        super().__init__(pageType, storyInstance)
        self.scene = Scene.load(self.story.styles["scenarios"][self.pageType])
        self.board = Board()
        self.narrations = {"transcripts": [], "interactions": []}
        self.narrator = self.story.narrator
        self.narratorId = -1
        self.boardContentScript = None
        self.boardContentListScripts = []
        self.defaultInteraction = None
        self.hasImage = False
        self.enableImageMagnify = True if self.pageType == "classroom" else False

        if all(key in kwargs for key in ("voices", "board", "objects", "interactions")):
            voices = kwargs.get("voices", None)
            self.scene = Scene.load(kwargs["scene"]) if kwargs.get("scene", None) != None else self.scene
            self.board = Board.load(kwargs.get("board", None))
            self.objects = kwargs.get("objects", None)
            if self.board.content.caption != None:
                self.boardContentScript = Script(transcript=self.board.content.caption)
            self.hasImage = True if (self.board.rect != None or self.board.content.rect != None or len(self.board.contentList) > 0) else False
            if self.objects != None:
                self.actor, self.actorId, self.narrator, self.narratorId, _, _ = get_actors(self.objects)

            if kwargs.get("interactions", None) != None:
                for interaction in kwargs["interactions"]:
                    if isinstance(interaction, dict):
                        # interactions that involves posters, or non visible-actor-involved
                        if "actor" in interaction and "content" in interaction \
                            and (("onResult" in interaction and (interaction["onResult"][0] if isinstance(interaction["onResult"], list) else interaction["onResult"]) > 0) \
                            or ("onPoster" in interaction and (interaction["onPoster"][0] if isinstance(interaction["onPoster"], list) else interaction["onPoster"]) > 0) \
                            or (self.objects[interaction["actor"]]["name"] not in VISIBLE_ACTORS)):
                            actorId = interaction["actor"]
                            content = interaction["content"]
                            if isinstance(content, dict) and "text" in content: # and "voice" in content and content["voice"] > 0:
                                text = content["text"]
                                voiceId = content.get("voice", -1)
                                self.narrations["transcripts"].append(Script(
                                    sound = voices[voiceId]["sound"] if voiceId > 0 else None,
                                    transcript = MText(text),
                                    narrator = self.objects[actorId]["name"],
                                    languages = voices[voiceId]["languages"] if (voiceId > 0 and "languages" in voices[voiceId] and isinstance(voices[voiceId]["languages"], list)) else None
                                ))
                                self.narrations["interactions"].append(Interaction(
                                    duration = interaction.get("duration", ""),
                                    actorId = actorId,
                                    figure = interaction.get("figure", None),
                                    content = Content(
                                        popup = 4,
                                        text = MText(text),
                                        voice = len(self.narrations["transcripts"]) - 1 if voiceId > 0 else voiceId
                                    ),
                                    onPoster = interaction.get("onResult", None) or interaction.get("onPoster", None),
                                    type = interaction.get("type", None)
                                ))
                        # visible-actor-involved interactions
                        elif "actor" in interaction:
                            actorId = interaction["actor"]
                            if "content" in interaction:
                                content = interaction["content"]
                                if isinstance(content, dict) and "text" in content: # and "voice" in content and content["voice"] > 0:
                                    text = content["text"]
                                    voiceId = content.get("voice", -1)
                                    self.transcripts.append(Script(
                                        sound=voices[voiceId]["sound"] if voiceId > 0 else None,
                                        transcript=MText(text),
                                        narrator=self.objects[actorId]["name"],
                                        languages=voices[voiceId]["languages"] if (voiceId > 0 and isinstance(voices[voiceId].get("languages", None), list)) else None
                                    ))
                                    self.interactions.append(Interaction(
                                        duration = interaction.get("duration", ""),
                                        content = Content(
                                            popup = content.get("popup", None),
                                            text = MText(text),
                                            voice = len(self.transcripts) - 1 if voiceId > 0 else voiceId
                                        ),
                                        actorId = actorId,
                                        type = interaction.get("type", None)
                                    ))
                            if ("figure" in interaction and interaction.get("figure", -1) > -1) and "position" in interaction:
                                self.defaultInteraction = Interaction(
                                    figure = interaction["figure"],
                                    position = interaction["position"],
                                    transform = interaction.get("transform", None),
                                    actorId = actorId,
                                    type = "talk"
                                )

    def updateActorPosition(self, oldPositions, newPositions):
        super().updateActorPosition(oldPositions, newPositions)
        if self.defaultInteraction!=None:
            currentPositionKey = None
            for key in oldPositions:
                if oldPositions[key] == self.defaultInteraction.position:
                    currentPositionKey = key
                    break
            if currentPositionKey in newPositions:
                self.defaultInteraction.position = newPositions[currentPositionKey]

    def updateActorTransform(self, oldTransform, newTransform):
        super().updateActorTransform(oldTransform, newTransform)
        if self.defaultInteraction != None and self.defaultInteraction.transform == oldTransform:
                self.defaultInteraction.transform = newTransform

    def setActor(self, actor, postures=["smilesay", "-stand-"], **kwargs):
        if actor in VISIBLE_ACTORS:
            if self.actor != actor:
                update_visible_actor(self.objects, actor)
                [script.reset2basename(actor) for script in self.transcripts]
                self.actor = actor

            posture_list = (
                [postures]
                if isinstance(postures, (str, int))
                else [postures[0]] \
                    if (isinstance(postures, list) and isinstance(postures[0], int))
                    else postures
            )
            figure, position = self.story.getUserPostureIdAndPosition(actor, posture_list, keyScenario="")
            if isinstance(kwargs.get("position", None), list) and len(kwargs["position"]) == 2:
                position = kwargs["position"]
            scale = None
            if isinstance(kwargs.get("scale", None), (int, float)) and kwargs["scale"] > 0:
                scale = f"scale({kwargs['scale']})"
            self.defaultInteraction = Interaction(
                                figure = figure,
                                position = position,
                                transform = scale if scale != None else self.story.styles["transform"],
                                type = "talk",
                                actorId = self._getUserId(actor)
                            )
        elif actor == None:
            self.defaultInteraction = None

    def setDialog(self, text:Union[str, dict], alternativeText:Union[str, dict]=None, **kwargs):
        newTranscript = Script(
            sound = switch_to_basename(self.transcripts[0]["sound"]) \
                if (len(self.transcripts) > 0 and isinstance(self.transcripts[0], dict) \
                    and self.transcripts[0].sound != None and len(self.transcripts[0].sound) > 0) \
                else f"voice-{uuid.uuid4()}.mp3",
            transcript = MText(text),
            narrator = self.actor,
            languages = kwargs["languages"] if isinstance(kwargs.get("languages", None), list) else None,
            alternative = MText(alternativeText) if alternativeText != None else None
        )
            
        newInteraction = Interaction(
            duration = "",
            content = Content(
                popup = kwargs.get("popup", self.story.styles["popup"]),
                text = MText(text),
                voice = 0
            ),
            type = "talk",
            figure = -1, 
            actorId = self._getUserId(self.actor)
        )
        self.transcripts = [newTranscript] + self.transcripts[1:] if self.transcripts else [newTranscript]
        self.interactions = [newInteraction] + self.interactions[1:] if self.interactions else [newInteraction]

    def addDialog(self, text:Union[str, dict], alternativeText:Union[str, dict]=None, **kwargs):
        self.transcripts.append(Script(
            sound = f"voice-{uuid.uuid4()}.mp3",
            transcript = MText(text),
            narrator = self.actor,
            languages = kwargs["languages"] if isinstance(kwargs.get("languages", None), list) else None,
            alternative = MText(alternativeText) if alternativeText != None else None
        ))

        self.interactions.append(Interaction(
            duration="auto",
            content = Content(
                popup=kwargs.get("popup", self.story.styles["popup"]),
                text=MText(text),
                voice=len(self.transcripts) - 1
            ),
            figure=-1,
            actorId=self._getUserId(self.actor),
            type="talk"))

    def updateDialog(self, pos, text:Union[str, dict]=None, alternativeText:Union[str, dict]=None, popup=None):
        if pos < len(self.transcripts) and pos >= 0:
            self.transcripts[pos].transcript = MText(self.transcripts[pos].transcript).merge(text)
            if alternativeText != None:
                self.transcripts[pos].alternative = MText(self.transcripts[pos].alternative).merge(alternativeText)
            self.transcripts[pos].reset2basename()

            self.interactions[pos].content.text = MText(self.interactions[pos].content.text).merge(text)
            if popup != None:
                self.interactions[pos].popup = popup

    def removeDialog(self, pos):
        if pos < len(self.transcripts) and pos >= 0:
            self.transcripts.pop(pos)
            self.interactions.pop(pos)
            if pos < len(self.interactions):
                for i, interaction in enumerate(self.interactions[pos:]):
                    self.interactions[i].content.voice = interaction.content.voice - 1

    def addNarration(self, text:Union[str, dict], narrator=None, alternativeText:Union[str, dict]=None, sound=None, **kwargs):
        narrator = self.pageType if kwargs.get("soundReady", None)==True \
            else (narrator if narrator != None else self.narrator if self.narrator != None else self.actor)

        newSound = os.path.basename(sound) if sound!=None else None
        if sound!=None and kwargs.get("soundReady", None)==True:
            try:
                self.story._cosUploader.local2cos(sound, self.story.storyId, self.story.audioPath)
                newSound = self.story.getAudioPath(os.path.basename(sound))
            except Exception as e:
                print(f"Upload sound file failed for [{sound}]\n", e)

        self.narrations["transcripts"].append(Script(
            sound = f"voice-{uuid.uuid4()}.mp3" if sound == None else newSound,
            transcript = MText(text),
            narrator = narrator,
            alternative = MText(alternativeText) if alternativeText!=None else None,
            languages = kwargs["languages"] if isinstance(kwargs.get("languages", None), list) else None,
            soundReady=bool(kwargs.get("soundReady", False))
        ))

        self.narrations["interactions"].append(Interaction(
            duration="auto",
            content = Content(
                popup = 4,
                text = MText(text),
                voice = len(self.narrations["transcripts"]) - 1
            ),
            figure = -1 if narrator in VISIBLE_ACTORS else None,
            actorId = self._getUserId(narrator),
            type = "talk"))

    def updateNarration(self, pos, text:Union[str, dict]=None, narrator=None, alternativeText:Union[str, dict]=None, sound=None, **kwargs):
        if pos < len(self.narrations["transcripts"]) and pos >= 0 \
            and (text!=None or narrator!=None or alternativeText!=None or (sound!=None and bool(kwargs.get("soundReady", False)))):

            newNarrator = narrator if narrator != None else self.objects[self.narrations["interactions"][pos].actorId]["name"]

            if text != None:
                self.narrations["transcripts"][pos].transcript = MText(self.narrations["transcripts"][pos].transcript).merge(text)
                self.narrations["interactions"][pos].content.text = MText(self.narrations["interactions"][pos].content.text).merge(text)
            if alternativeText != None:
                self.narrations["transcripts"][pos].alternative = MText(self.narrations["transcripts"][pos].alternative).merge(alternativeText)

            if sound!=None and bool(kwargs.get("soundReady", False)):
                self.narrations["transcripts"][pos].sound = os.path.basename(sound)
                try:
                    self.story._cosUploader.local2cos(sound, self.story.storyId, self.story.audioPath)
                except Exception as e:
                    self.narrations["transcripts"][pos].sound = self.story.getAudioPath(os.path.basename(sound))
                    print(f"Upload sound file failed for [{sound}]\n", e)
                self.narrations["transcripts"][pos].narrator = self.pageType
                self.narrations["transcripts"][pos].soundReady = True
                self.narrations["interactions"][pos].actorId = self._getUserId(self.pageType)
                self.narrations["interactions"][pos].figure = None
            else:
                if self.narrations["transcripts"][pos].soundReady:
                    self.narrations["transcripts"][pos].sound = f"voice-{uuid.uuid4()}.mp3"
                    self.narrations["transcripts"][pos].soundReady = False
                    newNarrator = narrator if narrator != None else self.narrator if self.narrator != None else self.actor
                else:
                    self.narrations["transcripts"][pos].reset2basename(narrator)
                    newNarrator = narrator if narrator != None else self.objects[self.narrations["interactions"][pos].actorId]["name"]
                self.narrations["transcripts"][pos].narrator = newNarrator
                self.narrations["interactions"][pos].actorId = self._getUserId(newNarrator)
                self.narrations["interactions"][pos].figure = -1 if newNarrator in VISIBLE_ACTORS else None

    def removeNarration(self, pos):
        if pos < len(self.narrations["transcripts"]) and pos >= 0:
            self.narrations["transcripts"].pop(pos)
            self.narrations["interactions"].pop(pos)
            if pos < len(self.narrations["interactions"]):
                for i, interaction in enumerate(self.narrations["interactions"][pos:]):
                    self.narrations["interactions"][i].content.voice = interaction.content.voice - 1

    def setImage(self, source:Union[str, dict], rect:list=[0.2, 0.2, 400, 400], autoFit:bool=True, uploadToCos:bool=True, **kwargs):
        assert isinstance(rect, list) and len(rect) >= 4
        mSource = MText(source) if source!=None else None
        mCaption = MText(kwargs.get("caption", self.board.content.caption if self.board.content.caption != None else ""))
        if kwargs.get("sourceType", None) != "video" and mSource!=None:
            validUrl = False
            if is_valid_http_url(mSource.firstValidPair()[1]):
                width = rect[2]
                height = rect[3]
                validUrl = True
            else:
                width, height = get_image_size(mSource.firstValidPair()[1])
            assert width > 0 and height > 0

            if not validUrl:
                if autoFit:
                    rect = Page.fit_image_rect(rect, width, height)

                if uploadToCos:
                    if isinstance(mSource.data, dict):
                        for key in mSource.data:
                            mSource.data[key] = self.story.uploadImageToCos(mSource.data[key], applyHash=kwargs.get("applyHash", True))
                    else:
                        mSource.data = self.story.uploadImageToCos(mSource.data, applyHash=kwargs.get("applyHash", True))

        self.board = Board(
            content = Content(
                caption = mCaption,
                image = mSource if kwargs.get("sourceType", None) != "video" else None,
                src = mSource if kwargs.get("sourceType", None) == "video" else None,
                videoType = kwargs.get("videoType", None) if kwargs.get("sourceType", None) == "video" else None,
                fontSize = kwargs.get("fontSize", "24px") if mCaption.len() > 0 else None,
                fontColor = kwargs.get("fontSize", "white") if mCaption.len() > 0 else None,
                magnify = kwargs.get("magnify", self.enableImageMagnify),
                border = kwargs.get("borderStyle", self.story.styles["frame"]) if kwargs.get("border", self.enableImageMagnify) else None
            ),
            rect = rect
        )

        self.boardContentScript = Script(transcript=mCaption)
        self.hasImage = True

    def setVideo(self, source:Union[str, dict], autoFit:bool=True, rect:list=[0.1, 0.1, 640, 360], **kwargs):
        self.setImage(source=source, rect=rect, autoFit=autoFit, uploadToCos=False, sourceType="video", **kwargs)

    def addImage(self, source:Union[str, dict], rect:list=[0, 0, 1, 1], autoFit:bool=True, uploadToCos:bool=True, **kwargs):
        assert isinstance(rect, list) and len(rect) >= 4 and source!=None
        validUrl = False
        if is_valid_http_url(MText(source).firstValidPair()[1]):
            width = rect[2]
            height = rect[3]
            validUrl = True
        else:
            width, height = get_image_size(MText(source).firstValidPair()[1])
        assert width > 0 and height > 0
        mSource = MText(source) if source!=None else None
        mCaption = MText(kwargs.get("caption", ""))

        if not validUrl:
            if autoFit:
                rect = Page.fit_image_rect(rect, width, height)

            if uploadToCos:
                if isinstance(mSource.data, dict):
                    for key in mSource.data:
                        mSource.data[key] = self.story.uploadImageToCos(mSource.data[key], applyHash=kwargs.get("applyHash", True))
                else:
                    mSource.data = self.story.uploadImageToCos(mSource.data, applyHash=kwargs.get("applyHash", True))

        self.board.contentList.append(Content(
            caption = mCaption,
            image = mSource,
            rect = rect,
            fontSize = kwargs.get("fontSize", "24px") if mCaption.len() > 0 else None,
            fontColor = kwargs.get("fontSize", "white") if mCaption.len() > 0 else None,
            magnify = kwargs.get("magnify", self.enableImageMagnify),
            border = kwargs.get("borderStyle", self.story.styles["frame"]) \
                if kwargs.get("border", self.enableImageMagnify) \
                else None
            )
        )
        self.boardContentListScripts.append(Script(transcript=mCaption))

    def updateImage(self, pos:int, source:MText, rect:list=[0, 0, 1, 1], autoFit=True, uploadToCos=True, **kwargs):
        if pos < len(self.board.contentList) and pos >= 0:
            assert isinstance(rect, list) and len(rect) >= 4
            validUrl = False
            if is_valid_http_url(MText(source).firstValidPair()[1]):
                width = rect[2]
                height = rect[3]
                validUrl = True
            else:
                width, height = get_image_size(MText(source).firstValidPair()[1])
            assert width > 0 and height > 0
            mSource = MText(source) if source!=None else None
            mCaption = MText(kwargs.get("caption", \
                                        self.board.contentList[pos].caption \
                                            if self.board.contentList[pos].caption != None \
                                            else ""
                                        )
                            )
            if isinstance(self.board.contentList[pos].caption, MText):
                mCaption = self.board.contentList[pos].caption.merge(mCaption)

            if not validUrl:
                if autoFit:
                    rect = Page.fit_image_rect(rect, width, height)

                if uploadToCos:
                    if isinstance(mSource.data, dict):
                        for key in mSource.data:
                            mSource.data[key] = self.story.uploadImageToCos(mSource.data[key], applyHash=kwargs.get("applyHash", True))
                    else:
                        mSource.data = self.story.uploadImageToCos(mSource.data, applyHash=kwargs.get("applyHash", True))

            self.board.contentList[pos].image = self.board.contentList[pos].image.merge(mSource)
            self.board.contentList[pos].rect = rect
            self.board.contentList[pos].caption = mCaption
            self.board.contentList[pos].magnify = True if kwargs.get("magnify", self.enableImageMagnify) else False
            self.board.contentList[pos].border = kwargs.get("borderStyle", self.story.styles["frame"]) \
                if kwargs.get("border", self.enableImageMagnify) \
                else None

            self.boardContentListScripts[pos].transcript = mCaption

    def removeImage(self, pos:int):
        if pos < len(self.board.contentList) and pos >= 0:
            self.board.contentList.pop(pos)
            self.boardContentListScripts.pop(pos)

    def __audible_script_list__(self):
        return [self.transcripts, self.narrations["transcripts"]]

    def export(self, voiceOffset=0, pageId=0.0):
        outTranscripts = [transcript.export() for transcript in self.transcripts \
                         if isinstance(transcript, Script) and transcript.export() is not None]
        # only count those transcript-with-voice
        narrationOffset = len(outTranscripts)
        outTranscripts += [transcript.export() for transcript in (self.narrations["transcripts"] \
                            + [self.boardContentScript] + self.boardContentListScripts) \
                         if (isinstance(transcript, Script) and transcript.export() is not None)]

        tempInteractions = []
        if len(self.interactions) > 0:
            for interaction in self.interactions:
                updatedInteraction = interaction.merge(self.defaultInteraction) if self.defaultInteraction != None else interaction.copy()
                if self.hasImage:
                    updatedInteraction.content.popup = 4
                tempInteractions.append(updatedInteraction)
        else:
            if self.defaultInteraction != None:
                defaultInteraction = self.defaultInteraction.copy()
                defaultInteraction.type = "motion"
                tempInteractions.append(defaultInteraction)


        for interaction in self.narrations["interactions"]:
            tempNarration = interaction.copy()
            tempNarration.content.voice += narrationOffset
            tempInteractions.append(tempNarration)

        outInteractions = []
        for interaction in tempInteractions:
            if interaction.content.voice != None and interaction.content.voice > -1:
                interaction.content.voice += voiceOffset
            exported = interaction.export()
            if exported != None:
                outInteractions.append(exported)

        eventDict = {
            "id": pageId,
            "scene": self.scene.export(),
            "board": self.board.export(),
            "objects": copy.deepcopy(self.objects),
            "interactions": outInteractions
        }

        if self.duration is not None:
            eventDict["duration"] = self.duration

        if self.transition is not None:
            eventDict["transition"] = self.transition

        return {
            "voices": outTranscripts,
            "events": [eventDict]
        }


##### 黑板页面 #####
class BlackboardPage(BoardPage):
    def __init__(self, storyInstance, **kwargs):
        super().__init__("blackboard", storyInstance, **kwargs)

        if len(kwargs.get("source", "")) > 0:
            self.setImage(**kwargs)

        if kwargs.get("actor", -1) in VISIBLE_ACTORS:
            self.setActor(**kwargs)

    def addNarration(self, text, narrator=None, alternativeText=None, sound=None, **kwargs):
        super().addNarration(text, narrator, alternativeText, sound, **kwargs)
        if kwargs.get("onPoster", None) != None or kwargs.get("onResult", None) != None:
            onPosterId = kwargs.get("onPoster", None) or kwargs.get("onResult", None)
            if onPosterId > len(self.board.contentList) + (1): # onResult=1 is always reserved for board image
                print(f"{YELLOW}Warning{RESET}: onPoster/onResult is greater than available image count and is ignored.")
            else:
                self.narrations["interactions"][-1].onPoster = onPosterId

    def updateNarration(self, pos, text=None, narrator=None, alternativeText=None, sound=None, **kwargs):
        super().updateNarration(pos, text, narrator, alternativeText, sound, **kwargs)
        if kwargs.get("onPoster", None) != None or kwargs.get("onResult", None) != None:
            onPosterId = kwargs.get("onPoster", None) or kwargs.get("onResult", None)
            if onPosterId > len(self.board.contentList) + (1): # onResult=1 is always reserved for board image
                print(f"{YELLOW}Warning{RESET}: onPoster/onResult is greater than available image count and is ignored.")
            self.narrations["interactions"][pos].onResult = None
            self.narrations["interactions"][pos].onPoster = onPosterId
        elif "onPoster" in kwargs or "onResult" in kwargs:
            self.narrations["interactions"][pos].onPoster = None
            self.narrations["interactions"][pos].onResult = None


##### 封面页面 #####
class CoverPage(BoardPage):
    def __init__(self, storyInstance, **kwargs):
        super().__init__("cover", storyInstance, **kwargs)

        if kwargs.get("source", None) != None and len(kwargs["source"]) > 0:
            self.setImage(**kwargs)

    def setActor(self, **kwargs):
        pass
    
    def addDialog(self, **kwargs):
        pass

    def updateDialog(self, **kwargs):
        pass
    
    def removeDialog(self, **kwargs):
        pass


##### 教室页面 #####
class ClassroomPage(BoardPage):
    def __init__(self, storyInstance, **kwargs):
        super().__init__("classroom", storyInstance, **kwargs)

        if "actor" in kwargs and kwargs["actor"] in VISIBLE_ACTORS:
            self.setActor(**kwargs)
    
    def addNarration(self, text, narrator=None, alternativeText=None, sound=None, **kwargs):
        super().addNarration(text, narrator, alternativeText, sound, **kwargs)
        if kwargs.get("onPoster", None) != None or kwargs.get("onResult", None) != None:
            onPosterId = kwargs.get("onPoster", None) or kwargs.get("onResult", None)
            if onPosterId > len(self.board.contentList) + (1): # onResult=1 is always reserved for board image
                print(f"{YELLOW}Warning{RESET}: onPoster/onResult is greater than available image count and is ignored.")
            else:
                self.narrations["interactions"][-1].onPoster = onPosterId

    def updateNarration(self, pos, text=None, narrator=None, alternativeText=None, sound=None, **kwargs):
        super().updateNarration(pos, text, narrator, alternativeText, sound, **kwargs)
        if kwargs.get("onPoster", None) != None or kwargs.get("onResult", None) != None:
            onPosterId = kwargs.get("onPoster", None) or kwargs.get("onResult", None)
            if onPosterId > len(self.board.contentList) + (1): # onResult=1 is always reserved for board image
                print(f"{YELLOW}Warning{RESET}: onPoster/onResult is greater than available image count and is ignored.")
            self.narrations["interactions"][pos].onResult = None
            self.narrations["interactions"][pos].onPoster = onPosterId
        elif "onPoster" in kwargs or "onResult" in kwargs:
            self.narrations["interactions"][pos].onPoster = None
            self.narrations["interactions"][pos].onResult = None


import os
import uuid

from .storyprofiles import CHARACTER_VOICE_PROFILES
from .speechsynthesizer import speech_synthesizer

#Replace with your azure subscription and region
AZURE_SPEECH_KEY = os.environ.get('AZURE_SPEECH_KEY')
AZURE_SERVICE_REGION = os.environ.get('AZURE_SPEECH_REGION')

class VoiceSynthesizer:
    def __init__(self, azureSubscription=AZURE_SPEECH_KEY, azureRegion=AZURE_SERVICE_REGION, voiceProfiles=CHARACTER_VOICE_PROFILES):
        self.voiceProfiles = voiceProfiles
        self.subscription = azureSubscription
        self.region = azureRegion

    def _newAudioFileName(self, language=None):
        if language and language.lower() not in ('cn', 'zh-cn'):
            return f'voice-{str(uuid.uuid4())}.{language}.mp3'
        else:
            return f'voice-{str(uuid.uuid4())}.mp3'
            
    def _fixAudioFileName(self, fileName, language):
        if f'{language}.mp3' in '.'.join(fileName.split('.')[-2:]):
            return fileName
        else:
            return '.'.join(fileName.split('.')[:-1])+f'.{language}.mp3'            

    @staticmethod
    def correctPronunciation(text, language, correctionDict):
        if language in correctionDict:
            for key in correctionDict[language]:
                text = text.replace(key, correctionDict[key])
        return text
    
    def synthesizeFile(self, character, text, language, outputPath, fileName=None, stopSymbols=['|', '\n']):
        synthesizer = speech_synthesizer(speech_key=self.subscription, service_region=self.region)

        character = 'M' if character == '' else character
        synthesizer.set_voice(
            language,
            CHARACTER_VOICE_PROFILES[character][language]['voiceName'],
            **CHARACTER_VOICE_PROFILES[character][language]['kwargs'])
        if fileName == None or len(fileName) < 1:
            fileName = self._newAudioFileName(language)
            print(f'No input file_name, generate new file name as: {fileName}')
        elif language.lower() not in ('cn', 'zh-cn'):
            fileName = self._fixAudioFileName(fileName, language)
            print(f'Filename with language code: {fileName}')

        print('Original:', text)
        newText = text
        for symbol in stopSymbols:
            newText = newText.split(symbol)[0]
        print('Corrected:', newText)

        try:
            synthesizer.synthesize(newText, outputPath, fileName)
        except Exception as e:
            print(e)

        return {
            "fileName": fileName,
            "originalText": text,
            "correctedText": newText,
            "character": character,
        }


from qcloud_cos import CosConfig
from qcloud_cos import CosS3Client
from qcloud_cos import CosServiceError
from qcloud_cos import CosClientError

import os
import hashlib


class CosUploader:
    def __init__(
        self,
        service_root,
        cos_region,
        cos_secret_id,
        cos_secret_key,
        cos_bucket,
        cos_token=None,
        cos_scheme="https",
    ):
        self.service_root = service_root
        self._config = CosConfig(
            Region=cos_region,
            SecretId=cos_secret_id,
            SecretKey=cos_secret_key,
            Token=cos_token,
            Scheme=cos_scheme,
        )
        self._client = CosS3Client(self._config)
        self._bucket = cos_bucket
        self._region = cos_region

    def calculate_file_hash(self, file_path):
        hash_md5 = hashlib.md5()
        with open(file_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def get_file_name(self, file_path, applyHash=True):
        filename, file_extension = os.path.splitext(os.path.basename(file_path))
        if file_extension in [".svg", ".jpg", ".jpeg", ".png", ".gif", ".bmp"]:
            return filename + (("_" + self.calculate_file_hash(file_path)) if applyHash else "") + file_extension
        else:
            return os.path.basename(file_path)

    def local2cos(self, local_path, storyId, target_relative_path="test", applyHash=True):
        if target_relative_path is None:
            target_relative_path = ""
        elif len(target_relative_path) > 0:
            target_relative_path = (
                target_relative_path
                if target_relative_path[0] != "/"
                else target_relative_path[1:]
            )
        try:
            filename = os.path.join(
                target_relative_path,
                storyId,
                self.get_file_name(local_path, applyHash=applyHash),
            )
        except:
            print("file not found:", local_path)
            return None
        
        response = self._client.upload_file(
            Bucket=self._bucket, LocalFilePath=local_path, Key=filename
        )
        return "/"+filename
    
    def test2product(self, source):
        source_file = source[1:] if source.startswith("/") else source
        target_file = source_file
        if source_file.startswith("test/audios/") or source_file.startswith("test/posters/"):
            target_file = "story/" + source_file[len("test/"):]
        if source_file != target_file:
            print(f"copy from {source_file} to {target_file}")
            try:
                response = self._client.copy_object(
                    Bucket=self._bucket,
                    Key=target_file,
                    CopySource={
                        'Bucket': self._bucket, 
                        'Key': source_file, 
                        'Region': self._region
                    }
                )
            except Exception as e:
                print("copy test file to product path failed:", e)
        return "/"+target_file

    def copy2dest(self, source, dest):
        try:
            self._client.copy(
                Bucket=self._bucket,
                Key=dest,
                CopySource={
                    'Bucket': self._bucket, 
                    'Key': source, 
                    'Region': self._region
                }
            )
            print(f'file copied from {source} to {dest}')
        except Exception as e:
            print('Exception occured in copy2dest():\n', e)
            return source
        return dest
